def print_locus(stranded_locus,
                gene_counter,
                handles,
                logger=None,
                configuration=None):
    """
    Method that handles a single superlocus for printing.
    It also detects and flags/discard fragmentary loci.
    :param stranded_locus: the stranded locus to analyse
    :return:

    :param gene_counter: integer used to keep track of the gene count.
    :type gene_counter: int

    :param handles: list of lists of the handles used for printing.
    :type handles: [list]

    :param logger: logger

    :param configuration: configuration dictionary
    :type configuration: [MikadoConfiguration|DaijinConfiguration]
    """

    locus_metrics, locus_scores, locus_out = handles[0]
    sub_metrics, sub_scores, sub_out = handles[1]
    mono_metrics, mono_scores, mono_out = handles[2]

    if configuration is None:
        from ..configuration.configurator import load_and_validate_config
        configuration = load_and_validate_config(None)

    stranded_locus.logger = logger
    if sub_out is not None:  # Skip this section if no sub_out is defined
        sub_lines = stranded_locus.__str__(
            level="subloci",
            print_cds=not configuration.pick.run_options.exclude_cds)
        if sub_lines != '':
            print(sub_lines, file=sub_out)
        sub_metrics_rows = [_ for _ in stranded_locus.print_subloci_metrics()
                            if _ != {} and "tid" in _]
        sub_scores_rows = [_ for _ in stranded_locus.print_subloci_scores()
                           if _ != {} and "tid" in _]
        for row in sub_metrics_rows:
            print(*[row[key] for key in sub_metrics.fieldnames],
                  sep="\t", file=sub_metrics.handle)
        for row in sub_scores_rows:
            print(*[row[key] for key in sub_scores.fieldnames],
                  sep="\t", file=sub_scores.handle)
    if mono_out is not None:
        mono_lines = stranded_locus.__str__(
            level="monosubloci",
            print_cds=not configuration.pick.run_options.exclude_cds)
        if mono_lines != '':
            print(mono_lines, file=mono_out)
        mono_metrics_rows = [_ for _ in stranded_locus.print_monoholder_metrics()
                             if _ != {} and "tid" in _]
        mono_scores_rows = [_ for _ in stranded_locus.print_monoholder_scores()
                            if _ != {} and "tid" in _]
        for row in mono_metrics_rows:
            print(*[row[key] for key in mono_metrics.fieldnames],
                  sep="\t", file=mono_metrics.handle)
        for row in mono_scores_rows:
            print(*[row[key] for key in mono_scores.fieldnames],
                  sep="\t", file=mono_scores.handle)

    for locus in stranded_locus.loci:
        gene_counter += 1
        new_id = "{0}.{1}G{2}".format(
            configuration.pick.output_format.id_prefix,
            stranded_locus.chrom, gene_counter)
        stranded_locus.loci[locus].logger = logger
        stranded_locus.loci[locus].id = new_id

    locus_lines = stranded_locus.__str__(
        print_cds=not configuration.pick.run_options.exclude_cds,
        level="loci")

    locus_metrics_rows = [x for x in stranded_locus.print_loci_metrics()]
    locus_scores_rows = [x for x in stranded_locus.print_loci_scores()]

    if locus_lines:
        assert len(locus_metrics_rows) > 0
        print(locus_lines, file=locus_out)

    for row in locus_metrics_rows:
        print(*[row[key] for key in locus_metrics.fieldnames],
              sep="\t", file=locus_metrics.handle)
    for row in locus_scores_rows:
        print(*[row[key] for key in locus_scores.fieldnames],
              sep="\t", file=locus_scores.handle)
    # Necessary to flush out all the files
    [_.flush() for _ in handles if hasattr(_, "close")]
    return gene_counter



from ..loci import Superlocus
from functools import partial
from ..utilities import default_for_serialisation
import rapidjson as json
dumper = partial(json.dumps, default=default_for_serialisation)
import msgpack


def serialise_locus(stranded_loci: [Superlocus],
                    accumulator,
                    counter,
                    print_subloci=True,
                    print_cds=True,
                    print_monosubloci=True):

    loci = []
    subloci = []
    monoloci = []
    for stranded_locus in sorted(stranded_loci):
        assert all(locus.metrics_calculated is True for locus in stranded_locus.loci.values())
        assert all(locus.scores_calculated is True for locus in stranded_locus.loci.values())
        loci.append([json.dumps(locus.as_dict(), default=default_for_serialisation)
                     for locus in stranded_locus.loci.values()])

        if print_subloci is True:
            batch = []
            batch.append(stranded_locus.__str__(level="subloci", print_cds=print_cds))
            if not all(sublocus.scores_calculated is True for sublocus in stranded_locus.subloci):
                raise RuntimeError("The locus we are trying to serialise has not been processed correctly.")

            sub_metrics_rows = [_ for _ in stranded_locus.print_subloci_metrics() if _ != {} and "tid" in _]
            sub_scores_rows = [_ for _ in stranded_locus.print_subloci_scores() if _ != {} and "tid" in _]
            batch.append(sub_metrics_rows)
            batch.append(sub_scores_rows)
            subloci.append(batch)

        if print_monosubloci is True:
            batch = []
            batch.append(stranded_locus.__str__(level="monosubloci", print_cds=print_cds))
            mono_metrics_rows = [_ for _ in stranded_locus.print_monoholder_metrics() if _ is not None
                                and _ != {} and "tid" in _]
            mono_scores_rows = [_ for _ in stranded_locus.print_monoholder_scores() if _ is not None
                                and _ != {} and "tid" in _]
            batch.append(mono_metrics_rows)
            batch.append(mono_scores_rows)
            monoloci.append(batch)

    loci = msgpack.dumps(loci)
    subloci = msgpack.dumps(subloci)
    monoloci = msgpack.dumps(monoloci)
    if not stranded_loci:
        chrom = ""
        num_genes = 0
    else:
        chrom = stranded_loci[0].chrom
        num_genes = sum(len(slid.loci) for slid in stranded_loci)
    accumulator.append((counter, chrom, num_genes, loci, subloci, monoloci))
    return accumulator



import argparse
import sqlite3

import pandas as pd
import pyfaidx

from Mikado.parsers import parser_factory

__doc__ = """Script to generate a test case for Mikado from a run"""

from Mikado.utilities import to_region


def generate_in_argument(id_list):
    return ','.join(f"'{name}'" for name in id_list)


def main():
    parser = argparse.ArgumentParser(__doc__)
    parser.add_argument("-p", "--prepared", type=parser_factory, help="Mikado prepare output file")
    parser.add_argument("-db", "--db", help="Mikado populated database (after serialise)")
    parser.add_argument("-ref", "--reference", type=pyfaidx.Fasta, help="Fasta reference file")
    parser.add_argument("-r", "--region", type=str, help="Region of interest to extract")
    parser.add_argument("-w", "--window", type=int, default=0,
                        help="Window up/down stream of the gene start/end to extract")
    parser.add_argument("-o", "--output_prefix", type=str, help="Output files prefix", default="test")
    args = parser.parse_args()


    id_list = []
    in_region = False
    target_chrom, target_start, target_end = to_region(args.region)

    # Extract region from genome
    min_seq_pos = max(1, target_start - 1 - args.window)
    max_seq_pos = min(len(args.reference[target_chrom]), target_end + args.window)
    seq = args.reference.get_seq(target_chrom,
                                 min_seq_pos + 1,
                                 max_seq_pos).seq
    print(f">{args.region}\n{seq}", file=open(args.output_prefix+'.fasta', 'w'))

    # Extract transcripts from prepare
    with open(args.output_prefix+".gtf", 'w') as gtf_output:
        for row in args.prepared:
            if row.header is True:
                continue
            else:
                if target_chrom == row.chrom and row.start >= target_start and row.end <= target_end:
                    in_region = True
                    if row.is_transcript:
                        # Generate a list of transcripts and get their query.ID from db
                        id_list.append(row.id)
                    row.start -= min_seq_pos
                    row.end -= min_seq_pos
                    row.chrom = args.region
                    print(row, file=gtf_output)
                elif in_region and (row.start >= target_end or target_chrom != row.chrom):
                    break

    input_db = sqlite3.connect(database=args.db)
    output_db = sqlite3.connect(database=args.output_prefix+'.db')

    chroms = pd.read_sql_query(f"SELECT * FROM chrom", input_db)
    chroms.loc[(chroms.name == target_chrom), 'name'] = args.region
    chroms.to_sql("chrom", output_db, if_exists='replace')

    # Extract db junctions from region remembering to shift by the same amount the genome is 'shifted'
    junctions = pd.read_sql_query(f"SELECT junctions.* FROM junctions "
                                  f"JOIN chrom ON junctions.chrom_id == chrom.chrom_id "
                                  f"WHERE "
                                  f"junctions.start >= {target_start} "
                                  f"AND junctions.end <= {target_end} AND "
                                  f"chrom.name == \"{target_chrom}\"", input_db)

    # Transform junctions to location in the extracted region including the window
    junctions['start'] -= min_seq_pos
    junctions['end'] -= min_seq_pos
    junctions['junction_start'] -= min_seq_pos
    junctions['junction_end'] -= min_seq_pos

    # Store junctions
    junctions.to_sql('junctions', output_db, if_exists='replace')

    # Extract db hsp,hit from transcript list using query.ID list
    query = pd.read_sql_query(f"SELECT * FROM query WHERE query_name IN ({generate_in_argument(id_list)})", input_db)
    query.to_sql('query', output_db, if_exists='replace')

    hsp = pd.read_sql_query(f"SELECT hsp.* FROM hsp WHERE query_id IN ({generate_in_argument(query['query_id'])})",
                            input_db)
    hsp.to_sql('hsp', output_db, if_exists='replace')

    hit = pd.read_sql_query(f"SELECT hit.* FROM hit WHERE hit.query_id IN ({generate_in_argument(query['query_id'])})",
                            input_db)
    hit.to_sql('hit', output_db, if_exists='replace')

    # Extract orf from db
    orf = pd.read_sql_query(f"SELECT * from orf WHERE query_id IN ({generate_in_argument(query['query_id'])})",
                            input_db)
    orf.to_sql('orf', output_db, if_exists='replace')

    target = pd.read_sql_query(f"SELECT * from target WHERE target_id IN ({generate_in_argument(hit['target_id'])})",
                               input_db)
    target.to_sql('target', output_db, if_exists='replace')


if __name__ == "__main__":
    main()



import re
import shutil
import unittest

import os

from Mikado import create_default_logger
from Mikado.exceptions import CorruptIndex, InvalidParsingFormat
from Mikado.parsers import parser_factory
from Mikado.scales.reference_preparation import indexing
from Mikado.scales.assignment import Assigner
import pkg_resources
import tempfile

from Mikado.transcripts import Transcript
from Mikado.utilities.namespace import Namespace


class TestIndexing(unittest.TestCase):

    # def setUp(self) -> None:

    def test_not_existent(self):
        logger = create_default_logger("test_not_existent")
        with tempfile.NamedTemporaryFile() as foo, self.assertRaises(CorruptIndex):
            indexing.check_index(foo.name, logger)
        with tempfile.NamedTemporaryFile() as foo, self.assertRaises(CorruptIndex):
            namespace = Namespace(default=False)
            namespace.reference = open(foo.name)
            indexing.load_index(namespace, logger)

    def test_index_various(self):
        logger = create_default_logger("test_index_gff3")
        for fname in ("trinity.gff3", "trinity.gtf", "trinity.bed12",
                      "trinity.cDNA_match.gff3", "trinity.match_matchpart.gff3"):
            with self.subTest(fname=fname), tempfile.TemporaryDirectory() as out_folder:
                shutil.copy(pkg_resources.resource_filename("Mikado.tests", fname), out_folder)
                fhandle = os.path.join(out_folder, fname)
                index_name = "{}.midx".format(fhandle)
                indexing.create_index(parser_factory(fhandle), logger, index_name,
                                      ref_gff=(fname.endswith("gff3")))
                self.assertTrue(os.path.exists(index_name))
                self.assertGreater(os.stat(index_name).st_size, 0)
                # Now check the index has been created correctly
                indexing.check_index(index_name, logger)
                namespace = Namespace(default=False)
                namespace.reference = parser_factory(fhandle)
                _ = indexing.load_index(namespace, logger)
                # Now rebuild
                with self.assertLogs(logger, level="INFO") as cmo:
                    indexing.create_index(parser_factory(fhandle), logger, index_name,
                                          ref_gff=(fname.endswith("gff3")))
                self.assertTrue(any([re.search(r"Removing the old index", _.msg) for _ in cmo.records]),
                                cmo.records)

    def test_protein_coding(self):
        fname = pkg_resources.resource_filename("Mikado.tests", "reference.gff3")

    def test_fusion(self):

        t = Transcript()
        t.chrom, t.strand, t.start, t.end, t.id, t.parent = "Chr1", "+", 101, 1000, "foo.1", "foo"
        t.add_exons([(101, 500), (601, 800), (901, 1000)])
        t.finalize()
        t2 = Transcript()
        t2.chrom, t2.strand, t2.start, t2.end, t2.id, t2.parent = "Chr1", "+", 2001, 3000, "bar.1", "bar"
        t2.add_exons([(2001, 2500), (2601, 2800), (2901, 3000)])
        t2.finalize()

        t3 = Transcript()
        t3.chrom, t3.strand, t3.start, t3.end, t3.id, t3.parent = "Chr1", "+", 651, 2703, "faz.1", "faz"
        t3.add_exons([(651, 800), (901, 1300), (2230, 2500), (2601, 2703)])
        t3.finalize()

        logger = create_default_logger("test_fusion")
        with tempfile.TemporaryDirectory() as folder:
            with open(os.path.join(folder, "reference.gtf"), "wt") as reference:
                print(t.format("gtf"), file=reference)
                print(t2.format("gtf"), file=reference)
            self.assertTrue(os.path.exists(reference.name))
            _ = [_ for _ in parser_factory(reference.name)]
            try:
                indexing.create_index(parser_factory(reference.name), logger, "{}.midx".format(reference.name))
            except InvalidParsingFormat:
                self.assertFalse(True, "\n".join([line.rstrip() for line in open(reference.name)]))
            namespace = Namespace(default=False)
            namespace.out = os.path.join(folder, "out")
            for report in (False, True):
                with self.subTest(report=report):
                    namespace.report_fusions = report
                    assigner = Assigner("{}.midx".format(reference.name), args=namespace, printout_tmap=False)
                    result = assigner.get_best(t3)
                    if report:
                        self.assertTrue(len(result), 2)
                        self.assertTrue(result[0].ccode == ("f", "j"), str(result[0]))
                        self.assertTrue(result[1].ccode == ("f", "j"), str(result[1]))
                    else:
                        self.assertTrue(result.ccode == ("j",), str(result))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# coding: utf-8

"""
Unit tests for the scales library
"""

import unittest
from .. import loci, utilities
from .. import scales
from ..scales.assignment.assigner import Assigner
from ..scales.assignment.distributed import encode_result


class AssignerTest(unittest.TestCase):
    """
    This unit test has the purpose of testing the scales module of Mikado.py.
    """

    def test_encode_result(self):
        res = scales.resultstorer.ResultStorer(
            "t1", "g1", "=",
            "p1", "pg1", "2", "2",
            100, 100, 100,
            100, 100, 100,
            100, 100, 100,
            0, "chr1:100..10000")
        encoded = encode_result(res)
        self.assertIsInstance(encoded, dict)
        self.assertEqual(encoded["ref_id"], "t1")
        self.assertEqual(encoded["distance"], "0")
        self.assertEqual(encoded["location"], "chr1:100..10000")
        encoded = encode_result(10)
        self.assertEqual(encoded, 10)

    def test_get_f1(self):

        # __slots__ = ["ref_id", "ref_gene", "ccode",
        #              "tid", "gid",
        #              "tid_num_exons", "ref_num_exons",
        #              "n_prec", "n_recall", "n_f1",
        #              "j_prec", "j_recall", "j_f1",
        #              "e_prec", "e_recall", "e_f1",
        #              "distance"]

        result_perfect = scales.resultstorer.ResultStorer(
            "t1", "g1", "=",
            "p1", "pg1", "2", "2",
            100, 100, 100,
            100, 100, 100,
            100, 100, 100,
            0, "chr1:100..10000")

        result_perfect_j = scales.resultstorer.ResultStorer(
            "t1", "g1", "=",
            "p1", "pg1", "2", "2",
            80, 80, 80,
            100, 100, 100,
            0, 0, 0,
            0, "chr1:100..10000")

        # This does not make any sense, but it's only for the tests
        result_perfect_n = scales.resultstorer.ResultStorer(
            "t1", "g1", "j",
            "p1", "pg1", "2", "2",
            100, 100, 100,
            80, 80, 80,
            0, 0, 0,
            0, "chr1:100..10000")

        result_imperfect = scales.resultstorer.ResultStorer(
            "t1", "g1", "j",
            "p1", "pg1", "2", "2",
            80, 80, 80,
            80, 80, 80,
            0, 0, 0,
            0, "chr1:100..10000")

        result_near = scales.resultstorer.ResultStorer(
            "t1", "g1", "p",
            "p1", "pg1", "2", "2",
            0, 0, 0,
            0, 0, 0,
            0, 0, 0,
            10, "chr1:100..10000")

        result_middle = scales.resultstorer.ResultStorer(
            "t1", "g1", "p",
            "p1", "pg1", "2", "2",
            0, 0, 0,
            0, 0, 0,
            0, 0, 0,
            500, "chr1:100..10000")

        result_far = scales.resultstorer.ResultStorer(
            "t1", "g1", "p",
            "p1", "pg1", "2", "2",
            0, 0, 0,
            0, 0, 0,
            0, 0, 0,
            1000, "chr1:100..10000")

        result_x = scales.resultstorer.ResultStorer(
            "t1", "g1", "x",
            "p1", "pg1", "2", "2",
            100, 100, 100,
            100, 100, 100,
            100, 100, 100,
            1000, "chr1:100..10000")

        self.assertEqual(sorted(
            [result_perfect, result_imperfect, result_perfect_j, result_perfect_n],
            key=Assigner.get_f1, reverse=True),
            [result_perfect, result_perfect_j, result_perfect_n, result_imperfect]
        )

        self.assertEqual(sorted([result_perfect, result_far], key=Assigner.get_f1, reverse=True),
                         [result_perfect, result_far])

        self.assertEqual(sorted(
            [result_far, result_near, result_middle, result_imperfect, result_perfect],
            key=Assigner.get_f1, reverse=True),
            [result_perfect, result_imperfect, result_near, result_middle,  result_far]
        )

        self.assertEqual(sorted(
            [result_perfect, result_x],
            key=Assigner.get_f1, reverse=True),
            [result_perfect, result_x]
        )

    def test_self(self):
        reference = loci.Transcript()
        reference.start = 100
        reference.end = 2000
        reference.strand = "+"
        reference.chrom = "Chr1"
        reference.id = "G1.1"
        reference.parent = "G1"
        reference.add_exons([(100, 300), (500, 1000), (1500, 2000)])
        reference.finalize()

        result, _ = Assigner.compare(reference, reference)
        self.assertEqual(result.ccode, ("=",))
        self.assertEqual(result.n_f1, (100,))
        self.assertEqual(result.n_prec, (100,))
        self.assertEqual(result.n_recall, (100,))
        self.assertEqual(result.j_f1, (100,))
        self.assertEqual(result.j_prec, (100,))
        self.assertEqual(result.j_recall, (100,))

    def test_mono_self(self):
        reference = loci.Transcript()
        reference.start = 100
        reference.end = 1000
        reference.chrom = "Chr1"
        reference.id = "G1.1"
        reference.parent = "G1"
        reference.strand = "+"
        reference.exons = [(100, 1000)]
        reference.finalize()

        result, _ = Assigner.compare(reference, reference)
        self.assertEqual(result.ccode, ("_",))
        self.assertEqual(result.n_f1, (100,))
        self.assertEqual(result.n_prec, (100,))
        self.assertEqual(result.n_recall, (100,))
        self.assertEqual(result.j_f1, (100,))
        self.assertEqual(result.j_prec, (100,))
        self.assertEqual(result.j_recall, (100,))

    def test_equal(self):
        reference = loci.Transcript()
        reference.start = 100
        reference.end = 2000
        reference.strand = "+"
        reference.chrom = "Chr1"
        reference.id = "G1.1"
        reference.parent = "G1"
        reference.exons = [(100, 300), (500, 1000), (1500, 2000)]
        reference.finalize()

        prediction = loci.Transcript()
        prediction.start = 200
        prediction.end = 1800
        prediction.strand = "+"
        prediction.chrom = "Chr1"
        prediction.id = "P1.1"
        prediction.parent = "P1"
        prediction.exons = [(200, 300), (500, 1000), (1500, 1800)]
        prediction.finalize()

        result, _ = Assigner.compare(prediction, reference)
        self.assertEqual(result.ccode, ("=",))
        self.assertEqual(result.j_f1, (100,))
        self.assertEqual(result.j_prec, (100,))
        self.assertEqual(result.j_recall, (100,))
        self.assertAlmostEqual(result.n_prec[0], 100, delta=0.1)
        self.assertAlmostEqual(result.n_recall[0],
                               100 * (300 - 200 + 1 + 1000 - 500 + 1 + 1800 - 1500 + 1) / reference.cdna_length,
                               delta=0.1)

        prediction.strand = "-"
        result, _ = Assigner.compare(prediction, reference)
        self.assertEqual(result.ccode, ("x",))

    def test_mono_equal(self):
        reference = loci.Transcript()
        reference.start = 100
        reference.end = 1000
        reference.chrom = "Chr1"
        reference.id = "G1.1"
        reference.parent = "G1"
        reference.strand = "+"
        reference.exons = [(100, 1000)]
        reference.finalize()

        prediction = loci.Transcript()
        prediction.start = 105
        prediction.end = 995
        prediction.chrom = "Chr1"
        prediction.id = "G1.1"
        prediction.parent = "G1"
        prediction.strand = "+"
        prediction.exons = [(105, 995)]
        prediction.finalize()

        result, _ = Assigner.compare(prediction, reference)
        self.assertEqual(result.ccode, ("_",))
        self.assertEqual(result.j_f1, (100,))
        self.assertEqual(result.j_prec, (100,))
        self.assertEqual(result.j_recall, (100,))
        self.assertAlmostEqual(result.n_prec[0], 100, delta=0.1)
        self.assertAlmostEqual(result.n_recall[0], 100 * (995 - 105 + 1) / reference.cdna_length, delta=0.1)

        prediction.strand = "-"
        result, _ = Assigner.compare(prediction, reference)
        self.assertEqual(result.ccode, ("x",))

    def test_mono_semiequal(self):
        reference = loci.Transcript()
        reference.start = 100
        reference.end = 1000
        reference.chrom = "Chr1"
        reference.id = "G1.1"
        reference.parent = "G1"
        reference.strand = "+"
        reference.exons = [(100, 1000)]
        reference.finalize()

        prediction = loci.Transcript()
        prediction.start = 200
        prediction.end = 500
        prediction.chrom = "Chr1"
        prediction.id = "G1.1"
        prediction.parent = "G1"
        prediction.strand = "+"
        prediction.exons = [(200, 500)]
        prediction.finalize()

        result, _ = Assigner.compare(prediction, reference)
        self.assertEqual(result.ccode, ("c",))
        self.assertEqual(result.j_f1, (100,))
        self.assertEqual(result.j_prec, (100,))
        self.assertEqual(result.j_recall, (100,))
        self.assertAlmostEqual(result.n_prec[0], 100, delta=0.1)
        self.assertAlmostEqual(result.n_recall[0], 100 * (500 - 200 + 1) / reference.cdna_length,
                               delta=0.1)

        prediction.strand = "-"
        result, _ = Assigner.compare(prediction, reference)
        self.assertEqual(result.ccode, ("x",))

    def test_mono_overlap(self):

        """Test that two monoexonic overlapping genes get a m"""

        reference = loci.Transcript()
        reference.start = 100
        reference.end = 1000
        reference.chrom = "Chr1"
        reference.id = "G1.1"
        reference.parent = "G1"
        reference.strand = "+"
        reference.exons = [(100, 1000)]
        reference.finalize()

        prediction = loci.Transcript()
        prediction.start = 200
        prediction.end = 3000
        prediction.chrom = "Chr1"
        prediction.id = "G1.1"
        prediction.parent = "G1"
        prediction.strand = "+"
        prediction.exons = [(200, 3000)]
        prediction.finalize()

        result, _ = Assigner.compare(prediction, reference)
        self.assertEqual(result.ccode, ("m",))
        self.assertEqual(result.j_f1, (100,))
        self.assertEqual(result.j_prec, (100,))
        self.assertEqual(result.j_recall, (100,))
        self.assertAlmostEqual(result.n_prec[0], 100 * 801 / prediction.cdna_length, delta=0.1)
        self.assertAlmostEqual(result.n_recall[0], 100 * 801 / reference.cdna_length, delta=0.1)

    def test_contained(self):

        reference = loci.Transcript()
        reference.start = 100
        reference.end = 2000
        reference.strand = "+"
        reference.chrom = "Chr1"
        reference.id = "G1.1"
        reference.parent = "G1"
        reference.exons = [(100, 300), (500, 1000), (1500, 2000)]
        reference.finalize()

        prediction = loci.Transcript()
        prediction.start = 600
        prediction.end = 1800
        prediction.strand = "+"
        prediction.chrom = "Chr1"
        prediction.id = "P1.1"
        prediction.parent = "P1"
        prediction.exons = [(600, 1000), (1500, 1800)]
        prediction.finalize()

        result, _ = Assigner.compare(prediction, reference)
        self.assertEqual(result.ccode, ("c",))
        self.assertAlmostEqual(result.j_f1[0], (100 * (2 / 3),)[0], delta=0.1)
        self.assertAlmostEqual(result.j_prec, (100,))
        self.assertAlmostEqual(result.j_recall, (100 * 1 / 2,), delta=0.1)
        self.assertAlmostEqual(result.n_prec[0], 100, delta=0.1)

        prediction.strand = "-"
        result, _ = Assigner.compare(prediction, reference)
        self.assertEqual(result.ccode, ("x",))

    def test_exon_skipping(self):

        """Test to verify that we call correctly an exon skipping as j"""

        reference = loci.Transcript()
        reference.chrom = 12
        reference.strand = "+"
        reference.add_exons([(7014003, 7014067),
                             (7014749, 7014923),
                             (7015573, 7015826),
                             (7016479, 7016609),
                             (7019054, 7019190),
                             (7021894, 7022191),
                             (7023055, 7023555),
                             (7015009, 7015118)  # This is our additional internal exon
                             ])
        reference.start = min(_[0] for _ in reference.exons)
        reference.end = max(_[1] for _ in reference.exons)
        reference.id = "t5"

        reference.finalize()

        prediction = loci.Transcript()
        prediction.chrom, prediction.strand = reference.chrom, reference.strand
        prediction.start, prediction.end = reference.start, reference.end
        prediction.add_exons([(7014003, 7014067),
                             (7014749, 7014923),
                             (7015573, 7015826),
                             (7016479, 7016609),
                             (7019054, 7019190),
                             (7021894, 7022191),
                             (7023055, 7023555)
                              ])
        prediction.finalize()
        prediction.id = "t3"

        self.assertEqual(len(reference.introns), 7)
        self.assertEqual(len(prediction.introns), 6)

        ref_vs_pred, _ = Assigner.compare(prediction, reference)
        self.assertEqual(ref_vs_pred.ccode, ("j", ), prediction.introns - reference.introns)
        pred_vs_ref, _ = Assigner.compare(reference, prediction)
        self.assertEqual(pred_vs_ref.ccode, ("j", ), prediction.introns - reference.introns)

    def test_alternative(self):

        reference = loci.Transcript()
        reference.start = 100
        reference.end = 2000
        reference.strand = "+"
        reference.chrom = "Chr1"
        reference.id = "G1.1"
        reference.parent = "G1"
        reference.exons = [(100, 300), (500, 1000), (1500, 2000)]
        reference.finalize()

        prediction = loci.Transcript()
        prediction.start = 100
        prediction.end = 2000
        prediction.strand = "+"
        prediction.chrom = "Chr1"
        prediction.id = "P1.1"
        prediction.parent = "P1"
        prediction.exons = [(100, 300), (600, 1000), (1500, 2000)]
        prediction.finalize()

        result, _ = Assigner.compare(prediction, reference)
        self.assertEqual(result.ccode, ("j",))
        self.assertAlmostEqual(result.j_f1[0], (100 * ((3 / 4) / 1),)[0], delta=0.1)
        self.assertAlmostEqual(result.j_prec, (100 * 3 / 4,), delta=0.1)
        self.assertAlmostEqual(result.j_recall, (100 * 3 / 4,), delta=0.1)

        prediction.strand = "-"
        result, _ = Assigner.compare(prediction, reference)
        self.assertEqual(result.ccode, ("X",))

    def test_mono_intronic(self):

        """
        R   |xxxxx|-------------------|xxxxxxxxx|
        P            |xxxxxxx|
        
        Expected class code: i
        """

        reference = loci.Transcript()
        reference.start = 100
        reference.end = 2000
        reference.strand = "+"
        reference.chrom = "Chr1"
        reference.id = "G1.1"
        reference.parent = "G1"
        reference.exons = [(100, 300), (500, 1000), (1500, 2000)]
        reference.finalize()

        prediction = loci.Transcript()
        prediction.start = 1100
        prediction.end = 1400
        prediction.strand = "+"
        prediction.chrom = "Chr1"
        prediction.id = "P1.1"
        prediction.parent = "P1"
        prediction.exons = [(1100, 1400)]
        prediction.finalize()

        result, _ = Assigner.compare(prediction, reference)
        self.assertEqual(result.ccode, ("i",))
        self.assertEqual(result.j_f1, (0,))
        self.assertEqual(result.j_prec, (0,))
        self.assertEqual(result.j_recall, (0,))
        if result.j_f1[0] > 0:
            self.assertAlmostEqual(result.j_f1[0], 2 * (result.j_recall[0] * result.j_prec[0]) /
                                   (result.j_prec[0] + result.j_recall[0]), delta=0.1)
        self.assertEqual(result.n_prec, (0,))
        self.assertEqual(result.n_recall, (0,))

        prediction.strand = "-"
        result, _ = Assigner.compare(prediction, reference)
        self.assertEqual(result.ccode, ("i",))

    def test_multi_intronic(self):

        """
        R   |xxxxx|-------------------|xxxxxxxxx|
        P            |xxx|----|xxxx|
        
        OR
        
        R   |xxxxx|----------|xxxxxxxxx|---------|xxxxxxxxx|
        P            |xxx|----------------|xxxx|
        
        Expected class code: I
        """

        reference = loci.Transcript()
        reference.start = 100
        reference.end = 2000
        reference.strand = "+"
        reference.chrom = "Chr1"
        reference.id = "G1.1"
        reference.parent = "G1"
        reference.exons = [(100, 300), (500, 1000), (1500, 2000)]
        reference.finalize()

        prediction = loci.Transcript()
        prediction.start = 1100
        prediction.end = 1400
        prediction.strand = "+"
        prediction.chrom = "Chr1"
        prediction.id = "P1.1"
        prediction.parent = "P1"
        prediction.exons = [(1100, 1200), (1300, 1400)]
        prediction.finalize()

        result, _ = Assigner.compare(prediction, reference)
        self.assertEqual(result.ccode, ("I",))
        self.assertEqual(result.j_f1, (0,))
        self.assertEqual(result.j_prec, (0,))
        self.assertEqual(result.j_recall, (0,))
        if result.j_f1[0] > 0:
            self.assertAlmostEqual(result.j_f1, 2 * (result.j_recall[0] * result.j_prec[0]) / (
                result.j_prec[0] + result.j_recall[0]), delta=0.1)
        self.assertEqual(result.n_prec, (0,))
        self.assertEqual(result.n_recall, (0,))

        prediction.strand = "-"
        result, _ = Assigner.compare(prediction, reference)
        self.assertEqual(result.ccode, ("I",))

        # Now the reference spans two introns
        prediction = loci.Transcript()
        prediction.start = 350
        prediction.end = 1400
        prediction.strand = "+"
        prediction.chrom = "Chr1"
        prediction.id = "P1.1"
        prediction.parent = "P1"
        prediction.exons = [(350, 450), (1300, 1400)]
        prediction.finalize()

        result, _ = Assigner.compare(prediction, reference)
        self.assertEqual(result.ccode, ("I",))
        self.assertEqual(result.j_f1, (0,))
        self.assertEqual(result.j_prec, (0,))
        self.assertEqual(result.j_recall, (0,))
        if result.j_f1[0] > 0:
            self.assertAlmostEqual(result.j_f1, 2 * (result.j_recall[0] * result.j_prec[0]) / (
                result.j_prec[0] + result.j_recall[0]), delta=0.1)
        self.assertEqual(result.n_prec, (0,))
        self.assertEqual(result.n_recall, (0,))

        prediction.strand = "-"
        result, _ = Assigner.compare(prediction, reference)
        self.assertEqual(result.ccode, ("I",))

    def test_overlap(self):

        """
        R   |xxxxx|-------|xxxxx|
        P      |xxxxx|---|xxxx|
        
        No junction in common
        
        """

        reference = loci.Transcript()
        reference.start = 100
        reference.end = 2000
        reference.strand = "+"
        reference.chrom = "Chr1"
        reference.id = "G1.1"
        reference.parent = "G1"
        reference.exons = [(100, 300), (500, 1000), (1500, 2000)]
        reference.finalize()

        prediction = loci.Transcript()
        prediction.start = 200
        prediction.end = 1300
        prediction.strand = "+"
        prediction.chrom = "Chr1"
        prediction.id = "P1.1"
        prediction.parent = "P1"
        prediction.exons = [(200, 700), (900, 1300)]
        prediction.finalize()

        result, _ = Assigner.compare(prediction, reference)
        self.assertEqual(result.ccode, ("o",))
        self.assertEqual(result.j_f1, (0,))
        self.assertEqual(result.j_prec, (0,))
        self.assertEqual(result.j_recall, (0,))
        n_overlap = (300 - 200 + 1 + 700 - 500 + 1 + 1000 - 900 + 1)
        self.assertAlmostEqual(result.n_prec[0], 100 * n_overlap / prediction.cdna_length, delta=0.1)
        self.assertAlmostEqual(result.n_recall[0], 100 * n_overlap / reference.cdna_length, delta=0.1)
        n_f1 = 2 * (n_overlap ** 2 / (reference.cdna_length * prediction.cdna_length)) / (
            n_overlap / reference.cdna_length + n_overlap / prediction.cdna_length)
        self.assertAlmostEqual(result.n_f1[0], 100 * n_f1, delta=0.1)

        prediction.strand = "-"
        result, _ = Assigner.compare(prediction, reference)
        self.assertEqual(result.ccode, ("X",))

    def test_ccode_e(self):

        """Case:
        
        R ---|xxxxx|-------|xxxxxx|---
        P      |xxxxxx|
        
        Exonic and intronic overlap
        
        """

        reference = loci.Transcript()
        reference.start = 100
        reference.end = 2000
        reference.strand = "+"
        reference.chrom = "Chr1"
        reference.id = "G1.1"
        reference.parent = "G1"
        reference.exons = [(100, 300), (500, 1000), (1500, 2000)]
        reference.finalize()

        prediction = loci.Transcript()
        prediction.start = 200
        prediction.end = 400
        prediction.strand = "+"
        prediction.chrom = "Chr1"
        prediction.id = "P1.1"
        prediction.parent = "P1"
        prediction.exons = [(200, 400)]
        prediction.finalize()

        result, _ = Assigner.compare(prediction, reference)
        self.assertEqual(result.ccode, ("e",), result)
        self.assertEqual(result.j_f1, (0,))
        self.assertEqual(result.j_prec, (0,))
        self.assertEqual(result.j_recall, (0,))
        n_overlap = (300 - 200 + 1)
        self.assertAlmostEqual(result.n_prec[0], 100 * n_overlap / prediction.cdna_length, delta=0.1)
        self.assertAlmostEqual(result.n_recall[0], 100 * n_overlap / reference.cdna_length, delta=0.1)
        n_f1 = 2 * (n_overlap ** 2 / (reference.cdna_length * prediction.cdna_length)) / (
            n_overlap / reference.cdna_length + n_overlap / prediction.cdna_length)
        self.assertAlmostEqual(result.n_f1[0], 100 * n_f1, delta=0.1)

        prediction.strand = "-"
        result, _ = Assigner.compare(prediction, reference)
        self.assertEqual(result.ccode, ("x",))

    def test_not_ccode_e(self):

        """Case:
        
        R ---|xxxxx|-------|xxxxxx|---
        P |xxxxxx|
        
        Exonic overlap only
        
        """
        reference = loci.Transcript()
        reference.start = 100
        reference.end = 2000
        reference.strand = "+"
        reference.chrom = "Chr1"
        reference.id = "G1.1"
        reference.parent = "G1"
        reference.exons = [(100, 300), (500, 1000), (1500, 2000)]
        reference.finalize()

        prediction = loci.Transcript()
        prediction.start = 50
        prediction.end = 310
        prediction.strand = "+"
        prediction.chrom = "Chr1"
        prediction.id = "P1.1"
        prediction.parent = "P1"
        prediction.exons = [(50, 310)]
        prediction.finalize()

        result, _ = Assigner.compare(prediction, reference)
        self.assertEqual(result.ccode, ("g",))
        self.assertEqual(result.j_f1, (0,))
        self.assertEqual(result.j_prec, (0,))
        self.assertEqual(result.j_recall, (0,))
        n_overlap = (300 - 100 + 1)
        self.assertAlmostEqual(result.n_prec[0], 100 * n_overlap / prediction.cdna_length, delta=0.1)
        self.assertAlmostEqual(result.n_recall[0], 100 * n_overlap / reference.cdna_length, delta=0.1)
        n_f1 = 2 * (n_overlap ** 2 / (reference.cdna_length * prediction.cdna_length)) / (
            n_overlap / reference.cdna_length + n_overlap / prediction.cdna_length)
        self.assertAlmostEqual(result.n_f1[0], 100 * n_f1, delta=0.1)

        prediction.strand = "-"
        result, _ = Assigner.compare(prediction, reference)
        self.assertEqual(result.ccode, ("x",))

    def test_left_extension(self):

        """
        R   |xxxxxxx|-------|xxxxx|
        P     |xxxxx|-------|xxxx|-------|xxxx|
        
        Expected ccode: j
        
        """

        reference = loci.Transcript()
        reference.start = 100
        reference.end = 2000
        reference.strand = "+"
        reference.chrom = "Chr1"
        reference.id = "G1.1"
        reference.parent = "G1"
        reference.exons = [(100, 300), (500, 1000), (1500, 2000)]
        reference.finalize()

        prediction = loci.Transcript()
        prediction.start = 50
        prediction.end = 1800
        prediction.strand = "+"
        prediction.chrom = "Chr1"
        prediction.id = "P1.1"
        prediction.parent = "P1"
        prediction.exons = [(50, 150), (200, 300), (500, 1000), (1500, 1800)]
        prediction.finalize()

        result, _ = Assigner.compare(prediction, reference)
        self.assertEqual(result.ccode, ("J",))
        self.assertEqual(result.j_f1, (4 / 5 * 100,))
        self.assertAlmostEqual(result.j_prec[0], 2 / 3 * 100, delta=0.1)
        self.assertEqual(result.j_recall, (100,))

    def test_left_extension_n(self):

        """
        R   |xxxxxxx|-------|xxxxx|
        P     |xxxxx|-------|xxxxx|-------|xxxx|

        Expected ccode: j

        """

        reference = loci.Transcript()
        reference.start = 100
        reference.end = 2000
        reference.strand = "+"
        reference.chrom = "Chr1"
        reference.id = "G1.1"
        reference.parent = "G1"
        reference.exons = [(100, 300), (500, 1000), (1500, 2000)]
        reference.finalize()

        prediction = loci.Transcript()
        prediction.start = 200
        prediction.end = 3000
        prediction.strand = "+"
        prediction.chrom = "Chr1"
        prediction.id = "P1.1"
        prediction.parent = "P1"
        prediction.exons = [(200, 300), (500, 1000), (1500, 2050), (2200, 3000)]
        prediction.finalize()

        result, _ = Assigner.compare(prediction, reference)
        self.assertEqual(result.ccode, ("n",))
        self.assertEqual(result.j_f1, (4 / 5 * 100,))
        self.assertAlmostEqual(result.j_prec[0], 2 / 3 * 100, delta=0.1)
        self.assertEqual(result.j_recall, (100,))

    def test_right_extension(self):
        """
        R                |xxxxxxx|-------|xxxxx|
        P     |xxxxx|-------|xxxx|-------|xxxx|
        
        Expected ccode: j
        """

        reference = loci.Transcript()
        reference.start = 100
        reference.end = 2000
        reference.strand = "+"
        reference.chrom = "Chr1"
        reference.id = "G1.1"
        reference.parent = "G1"
        reference.exons = [(100, 300), (500, 1000), (1500, 2000)]
        reference.finalize()

        prediction = loci.Transcript()
        prediction.start = 200
        prediction.end = 3000
        prediction.strand = "+"
        prediction.chrom = "Chr1"
        prediction.id = "P1.1"
        prediction.parent = "P1"
        prediction.exons = [(200, 300), (500, 1000), (1500, 1800), (2500, 3000)]
        prediction.finalize()

        result, _ = Assigner.compare(prediction, reference)
        self.assertEqual(result.ccode, ("J",))
        self.assertEqual(result.j_f1, (4 / 5 * 100,))
        self.assertAlmostEqual(result.j_prec[0], 2 / 3 * 100, delta=0.1)
        self.assertEqual(result.j_recall, (100,))

    def test_left_right_extension(self):
        """
        R                |xxxxxxx|-------|xxxxx|
        P     |xxxxx|-------|xxxx|-------|xxxx|------|xxxxx|
        
        Expected ccode: j
        """

        reference = loci.Transcript()
        reference.start = 100
        reference.end = 2000
        reference.strand = "+"
        reference.chrom = "Chr1"
        reference.id = "G1.1"
        reference.parent = "G1"
        reference.exons = [(100, 300), (500, 1000), (1500, 2000)]
        reference.finalize()

        prediction = loci.Transcript()
        prediction.start = 50
        prediction.end = 3000
        prediction.strand = "+"
        prediction.chrom = "Chr1"
        prediction.id = "P1.1"
        prediction.parent = "P1"
        prediction.exons = [(50, 150), (200, 300), (500, 1000), (1500, 1800), (2500, 3000)]
        prediction.finalize()

        result, _ = Assigner.compare(prediction, reference)
        self.assertEqual(result.ccode, ("J",))
        self.assertAlmostEqual(result.j_f1[0], 2 / 3 * 100, delta=0.1)
        self.assertAlmostEqual(result.j_prec[0], 1 / 2 * 100, delta=0.1)
        self.assertEqual(result.j_recall, (100,))

    def test_left_right_extension_novel(self):
        """
        R                   |xxxx|-------|xxxx|
        P     |xxxxx|----|xxxxxxx|-------|xxxxx|------|xxxxx|

        Expected ccode: n
        """

        reference = loci.Transcript()
        reference.start = 1000
        reference.end = 3000
        reference.strand = "+"
        reference.chrom = "Chr1"
        reference.id = "G1.1"
        reference.parent = "G1"
        reference.exons = [(1000, 1300), (1500, 2000), (2500, 3000)]
        reference.finalize()

        prediction = loci.Transcript()
        prediction.start = 500
        prediction.end = 3500
        prediction.strand = "+"
        prediction.chrom = "Chr1"
        prediction.id = "P1.1"
        prediction.parent = "P1"
        prediction.exons = [(500, 700), (900, 1300), (1500, 2000), (2500, 3100), (3200, 3500)]
        prediction.finalize()

        result, _ = Assigner.compare(prediction, reference)
        self.assertEqual(result.ccode, ("n",))
        self.assertAlmostEqual(result.j_f1[0], 2 / 3 * 100, delta=0.1)
        self.assertAlmostEqual(result.j_prec[0], 1 / 2 * 100, delta=0.1)
        self.assertEqual(result.j_recall, (100,))

    def test_internal_extension(self):
        """
        
        R    |xxxxxxxx|-----------------|xxxxxxxx|
        P       |xxxxx|----|xxxxxx|-----|xxxxx|
        
        Expected ccode: j, junction recall: 100%
        """

        reference = loci.Transcript()
        reference.start = 100
        reference.end = 2000
        reference.strand = "+"
        reference.chrom = "Chr1"
        reference.id = "G1.1"
        reference.parent = "G1"
        reference.exons = [(100, 300), (500, 1000), (1500, 2000)]
        reference.finalize()

        prediction = loci.Transcript()
        prediction.start = 200
        prediction.end = 2100
        prediction.strand = "+"
        prediction.chrom = "Chr1"
        prediction.id = "P1.1"
        prediction.parent = "P1"
        prediction.exons = [(200, 300), (500, 1000), (1200, 1300), (1500, 2100)]
        prediction.finalize()

        result, _ = Assigner.compare(prediction, reference)
        self.assertEqual(result.ccode, ("j",))
        self.assertAlmostEqual(result.j_f1[0], 4 / 5 * 100, delta=0.1)
        self.assertAlmostEqual(result.j_prec[0], 2 / 3 * 100, delta=0.1)
        self.assertEqual(result.j_recall, (100,))

    def test_internal_external_extension(self):
        """
        
        R      |xxxxxx|-----------------|xxxxxxxx|
        P       |xxxxx|----|xxxxxx|-----|xxxxx|------|xxxxxxx|
        
        Expected ccode: j, junction recall: 100%
        """

        reference = loci.Transcript()
        reference.start = 100
        reference.end = 2000
        reference.strand = "+"
        reference.chrom = "Chr1"
        reference.id = "G1.1"
        reference.parent = "G1"
        reference.exons = [(100, 300), (500, 1000), (1500, 2000)]
        reference.finalize()

        prediction = loci.Transcript()
        prediction.start = 200
        prediction.end = 3000
        prediction.strand = "+"
        prediction.chrom = "Chr1"
        prediction.id = "P1.1"
        prediction.parent = "P1"
        prediction.exons = [(200, 300), (500, 1000), (1200, 1300), (1500, 1800), (2500, 3000)]
        prediction.finalize()

        result, _ = Assigner.compare(prediction, reference)
        self.assertEqual(result.ccode, ("j",))
        self.assertAlmostEqual(result.j_f1[0], 2 / 3 * 100, delta=0.1)
        self.assertAlmostEqual(result.j_prec[0], 1 / 2 * 100, delta=0.1)
        self.assertEqual(result.j_recall, (100,))

    def test_left_right_internal_extension_novel(self):
        """
        R                   |xxxx|-------|xxxx|
        P     |xxxxx|----|xxxxxxx|--|x|--|xxxxx|------|xxxxx|

        Expected ccode: j
        """

        reference = loci.Transcript()
        reference.start = 1000
        reference.end = 3000
        reference.strand = "+"
        reference.chrom = "Chr1"
        reference.id = "G1.1"
        reference.parent = "G1"
        reference.exons = [(1000, 1300), (1500, 2000), (2500, 3000)]
        reference.finalize()

        prediction = loci.Transcript()
        prediction.start = 500
        prediction.end = 3500
        prediction.strand = "+"
        prediction.chrom = "Chr1"
        prediction.id = "P1.1"
        prediction.parent = "P1"
        prediction.exons = [(500, 700), (900, 1300), (1500, 2000),
                            (2200,2300),
                            (2500, 3100), (3200, 3500)]
        prediction.finalize()

        result, _ = Assigner.compare(prediction, reference)
        self.assertEqual(result.ccode, ("j",))
        self.assertAlmostEqual(result.j_f1[0], 2 * 0.4 / 1.4 * 100, delta=0.1)
        self.assertAlmostEqual(result.j_prec[0], 2 / 5 * 100, delta=0.1)
        self.assertEqual(result.j_recall, (100,))

    def test_contained_bleeding(self):

        """
        R     |xxxxxxx|--------|xxxxxx|----|xxxx|--------|xxxxxxx|
        P                  |xxxxxxxxxx|----|xxxxxx|

        Expected class code: C
        :return:
        """
        reference = loci.Transcript()
        reference.start = 1000
        reference.end = 3000
        reference.strand = "+"
        reference.chrom = "Chr1"
        reference.id = "G1.1"
        reference.parent = "G1"
        reference.exons = [(1000, 1300), (1500, 2000), (2200, 2400), (2500, 3000)]
        reference.finalize()

        prediction = loci.Transcript()
        prediction.start = 1400
        prediction.end = 2450
        prediction.strand = "+"
        prediction.chrom = "Chr1"
        prediction.id = "P1.1"
        prediction.parent = "P1"
        prediction.exons = [(1400, 2000), (2200, 2450)]
        prediction.finalize()

        result, _ = Assigner.compare(prediction, reference)
        self.assertEqual(result.ccode, ("C",), result)

    def test_contained_alternative(self):

        """
        R     |xxxxxxx|--------|xxxxxx|----|xxxx|--------|xxxxxxx|
        P           |xxxxxxxxxxxxxxxxx|----|xxxxxx|

        Expected class code: C
        :return:
        """
        reference = loci.Transcript()
        reference.start = 1000
        reference.end = 3000
        reference.strand = "+"
        reference.chrom = "Chr1"
        reference.id = "G1.1"
        reference.parent = "G1"
        reference.exons = [(1000, 1300), (1500, 2000), (2200, 2400), (2500, 3000)]
        reference.finalize()

        prediction = loci.Transcript()
        prediction.start = 1200
        prediction.end = 2450
        prediction.strand = "+"
        prediction.chrom = "Chr1"
        prediction.id = "P1.1"
        prediction.parent = "P1"
        prediction.exons = [(1200, 2000), (2200, 2450)]
        prediction.finalize()

        result, _ = Assigner.compare(prediction, reference)
        self.assertEqual(result.ccode, ("j",))

    def test_mono_overlap_nostrand(self):

        reference = loci.Transcript()
        reference.start = 100
        reference.end = 2000
        reference.strand = "-"
        reference.chrom = "Chr1"
        reference.id = "G1.1"
        reference.parent = "G1"
        reference.exons = [(100, 2000)]
        reference.finalize()

        prediction = loci.Transcript()
        prediction.start = 300
        prediction.end = 3000
        prediction.strand = "."
        prediction.chrom = "Chr1"
        prediction.id = "G1.1"
        prediction.parent = "G1"
        prediction.exons = [(300, 3000)]
        prediction.finalize()

        result, _ = Assigner.compare(prediction, reference)
        self.assertEqual(result.ccode, ("m",))

        result, _ = Assigner.compare(reference, prediction)
        self.assertEqual(result.ccode, ("m",))

    def test_mono_multi_overlap_nostrand(self):

        """Test a monoexonic transcript against a multiexonic transcript
         to verify the assignment of correct ccodes (o and O)
        """

        reference = loci.Transcript()
        reference.start = 100
        reference.end = 2000
        reference.strand = "+"
        reference.chrom = "Chr1"
        reference.id = "G1.1"
        reference.parent = "G1"
        reference.exons = [(100, 700), (1000, 2000)]
        reference.finalize()

        prediction = loci.Transcript()
        prediction.start = 50
        prediction.end = 600
        prediction.strand = "."
        prediction.chrom = "Chr1"
        prediction.id = "G1.1"
        prediction.parent = "G1"
        prediction.exons = [(50, 600)]
        prediction.finalize()

        result, _ = Assigner.compare(prediction, reference)
        self.assertEqual(result.ccode, ("g",))

        result, _ = Assigner.compare(reference, prediction)
        self.assertEqual(result.ccode, ("G",))

    def test_neighbors(self):

        """
        Test for the assignment of transcripts inside the index. Still a stub
        """
        keys = [(10, 200), (350, 500)]
        keys = utilities.intervaltree.IntervalTree.from_tuples(keys)
        self.assertEqual(Assigner.find_neighbours(
                         keys, (350,500), distance=2000),
                         [((350, 500), 0), ((10, 200), 150)],
                         )
        self.assertEqual(Assigner.find_neighbours(
                         keys, (5350,5500), distance=1000),
                         []
                         )

        self.assertEqual(Assigner.find_neighbours(
                         keys, (5350,5500), distance=10000),
                         [((350, 500), 4850), ((10, 200), 5150)]
                         )

    def test_monoexonic_contained(self):
        t1 = loci.Transcript()
        t1.chrom, t1.start, t1.end, t1.strand, t1.id = "Chr1", 9192977, 9193471, "+", "t1"
        t1.add_exons([(t1.start, t1.end)])
        t1.add_exons([(t1.start, t1.end)], features=["CDS"])
        t1.finalize()
        t2 = loci.Transcript()
        t2.chrom, t2.start, t2.end, t2.strand, t2.id = "Chr1", 9182621, 9194526, "+", "t2"
        t2.add_exons([(9192977, 9193480)], features=["CDS"])
        t2.add_exons([(9186643, 9186780), (9182621, 9182770), (9192959, 9194526)])
        t2.finalize()
        mono_ref = Assigner.compare(reference=t1, prediction=t2)[0]
        mono_pred = Assigner.compare(reference=t2, prediction=t1)[0]
        self.assertEqual(mono_ref.n_recall, mono_pred.n_prec)
        self.assertEqual(mono_pred.ccode[0], "c")
        self.assertEqual(mono_ref.ccode[0], "n")

    def test_h_case(self):

        """
        |xxxxxxxxxxxxxxx|-------------------|xxxxxxxxxxxx|
        |xxxxxxxxxxxxxxxxxxx|-----------------|xxxxxxxxx|
        :return:
        """

        reference = loci.Transcript()
        reference.start = 1000
        reference.end = 3000
        reference.strand = "+"
        reference.chrom = "Chr1"
        reference.id = "G1.1"
        reference.parent = "G1"
        reference.exons = [(1000, 1300), (2500, 3000)]
        reference.finalize()

        prediction = loci.Transcript()
        prediction.start = 1200
        prediction.end = 2450
        prediction.strand = "+"
        prediction.chrom = "Chr1"
        prediction.id = "P1.1"
        prediction.parent = "P1"
        prediction.exons = [(1200, 2000), (2200, 2450)]
        prediction.finalize()

        result, _ = Assigner.compare(prediction, reference)
        self.assertEqual(result.ccode, ("h",))

    def test_double_h_case(self):

        """
        |xxxxxxxxxxxxxxx|---------|xxxxxxx|-------|xxxxxxxxxxxx|
        |xxxxxxxxxxxxxxxxxxx|-----------------|xxxxxxxxx|
        :return:
        """

        reference = loci.Transcript()
        reference.start = 1000
        reference.end = 3000
        reference.strand = "+"
        reference.chrom = "Chr1"
        reference.id = "G1.1"
        reference.parent = "G1"
        reference.exons = [(1000, 1300), (1700, 2000), (2500, 3000)]
        reference.finalize()

        prediction = loci.Transcript()
        prediction.start = 1200
        prediction.end = 2450
        prediction.strand = "+"
        prediction.chrom = "Chr1"
        prediction.id = "P1.1"
        prediction.parent = "P1"
        prediction.exons = [(1200, 1500), (2200, 2450)]
        prediction.finalize()

    def test_non_h_case(self):

        """
        |xxxxxxxxxxxxxxx|-------------------|xxxxxxxxxxxx|
        |xxxxx|-----|xxxxxxxxx|
        :return:
        """

        reference = loci.Transcript()
        reference.start = 1000
        reference.end = 3000
        reference.strand = "+"
        reference.chrom = "Chr1"
        reference.id = "G1.1"
        reference.parent = "G1"
        reference.exons = [(1000, 1800), (2500, 3000)]
        reference.finalize()

        prediction = loci.Transcript()
        prediction.start = 1200
        prediction.end = 2050
        prediction.strand = "+"
        prediction.chrom = "Chr1"
        prediction.id = "P1.1"
        prediction.parent = "P1"
        prediction.exons = [(1200, 1600), (1700, 2050)]
        prediction.finalize()

        result, _ = Assigner.compare(prediction, reference)
        self.assertEqual(result.ccode, ("o",))

    def test_J_and_C_case(self):

        """

        1  xxxxxxxxx------------xxxxxxx-----------xxxxxxxx-------------xxxxxxxxxxxxx
        2                 xxxxxxxxxxxxx-----------xxxxxxxxxxxx

        We do expect the comparison to be:

        1 reference, 2 prediction: C
        1 prediction, 2 reference: J

        :return:
        """

        t1 = loci.Transcript()
        t1.chrom = "Chr1"
        t1.strand = "+"
        t1.score = 10
        t1.source = "mikado"
        t1.id = "t1.1"
        t1.parent = "t1"
        t1.start = 101
        t1.end = 5000
        t1.add_exons([(101, 1000), (1501, 2000), (2301, 4000), (4501, 5000)])
        t1.finalize()

        t2 = loci.Transcript()
        for attr in ["chrom", "source", "strand", "score"]:
            setattr(t2, attr, getattr(t1, attr))
        t2.id = "t2.1"
        t2.parent = "t2"
        t2.start = 1300
        t2.end = 4300
        t2.add_exons([(1300,2000), (2301, 4300)])
        t2.finalize()

        result, _ = Assigner.compare(t2, t1)
        self.assertEqual(result.ccode, ("C",))

        result, _ = Assigner.compare(t1, t2)
        self.assertEqual(result.ccode, ("J",))

    def test_J_and_C_case_in_exon(self):

        """

        1  xxxxxxxxx------------xxxxxxx-----------xxxxxxxx-------------xxxxxxxxxxxxx
        2         xxxxxxxxxxxxxxxxxxxxx-----------xxxxxxxxxxxx

        We do expect the comparison to be:

        1 reference, 2 prediction: j
        1 prediction, 2 reference: J

        Notice the class code switch from C to j due to the first exon being internal to the terminal exon,
        rather than internal to the intron.

        :return:
        """

        t1 = loci.Transcript()
        t1.chrom = "Chr1"
        t1.strand = "+"
        t1.score = 10
        t1.source = "mikado"
        t1.id = "t1.1"
        t1.parent = "t1"
        t1.start = 101
        t1.end = 5000
        t1.add_exons([(101, 1000), (1501, 2000), (2301, 4000), (4501, 5000)])
        t1.finalize()

        t2 = loci.Transcript()
        for attr in ["chrom", "source", "strand", "score"]:
            setattr(t2, attr, getattr(t1, attr))
        t2.id = "t2.1"
        t2.parent = "t2"
        t2.start = 900
        t2.end = 4300
        t2.add_exons([(900,2000), (2301, 4300)])
        t2.finalize()

        result, _ = Assigner.compare(t2, t1)
        self.assertEqual(result.ccode, ("j",))

        result, _ = Assigner.compare(t1, t2)
        self.assertEqual(result.ccode, ("J",))

    def test_fuzzy_match(self):

        t1 = loci.Transcript()
        t1.chrom = "Chr1"
        t1.strand = "+"
        t1.score = 10
        t1.source = "mikado"
        t1.id = "t1.1"
        t1.parent = "t1"
        t1.start = 101
        t1.end = 5000
        t1.add_exons([(101, 1000), (1501, 2000), (2301, 4000), (4501, 5000)])
        t1.finalize()

        t2 = loci.Transcript()
        t2.chrom = "Chr1"
        t2.strand = "+"
        t2.score = 10
        t2.source = "mikado"
        t2.id = "t2.1"
        t2.parent = "t2"
        t2.start = 401
        t2.end = 5130
        t2.add_exons([(401, 1010), (1491, 2005), (2305, 4002), (4509, 5130)])
        t2.finalize()

        for fuzzymatch in (0, 1, 5, 10, 20, 30):
            with self.subTest(fuzzymatch=fuzzymatch):
                result, _ = Assigner.compare(t2, t1, fuzzy_match=fuzzymatch)
                if fuzzymatch < 5:
                    self.assertEqual(result.ccode, ("h",))
                elif fuzzymatch >= 10:
                    self.assertEqual(result.ccode, ("=",))
                else:
                    self.assertEqual(result.ccode, ("j",), fuzzymatch)


if __name__ == '__main__':
    unittest.main()



"""
This module contains the methods used by the Transcript class to retrieve information
from the database/dictionary provided during the pick operation.
"""

import operator
from itertools import groupby
from typing import List
from sqlalchemy import and_
from sqlalchemy.orm.session import sessionmaker
from ...parsers.bed12 import BED12
from ...serializers.junction import Junction
from ..._transcripts.clique_methods import define_graph, find_cliques, find_communities
from ...utilities import dbutils

__author__ = 'Luca Venturini'


def load_orfs(transcript, candidate_orfs):

    """
    :param transcript: the Transcript instance
    :type transcript: Mikado.loci_objects.transcript.Transcript

    :param candidate_orfs: The ORFs to be inspected for loading.
    :type candidate_orfs: list[Mikado.serializers.orf.Orf|Mikado.parsers.bed12.BED12]

    This method replicates what is done internally by the
    "cdna_alignment_orf_to_genome_orf.pl"
    utility in the TransDecoder suite. It takes as argument "candidate_orfs"
    i.e. a list of BED12 serialised objects.
    The method expects as argument a dictionary containing BED entries,
    and indexed by the transcript name. The indexed name *must* equal the
    "id" property, otherwise the method returns immediately.
    If no entry is found for the transcript, the method exits immediately.
    Otherwise, any CDS information present in the original GFF/GTF
    file is completely overwritten.
    Briefly, it follows this logic:
    - Finalise the transcript
    - Retrieve from the dictionary (input) the CDS object
    - Sort in decreasing order the CDSs on the basis of:
        - Presence of start/stop codon
        - CDS length (useful for monoexonic transcripts where we might want to set the strand)
    - For each CDS:
        - If the ORF is on the + strand:
            - all good
        - If the ORF is on the - strand:
            - if the transcript is monoexonic: invert its genomic strand
            - if the transcript is multiexonic: skip
        - Start looking at the exons
    """

    # Prepare the transcript
    transcript.finalize()

    transcript.logger.debug("Finding the candidate ORFS out of %d .. ",
                            len(candidate_orfs))
    # This will also exclude invalid ORFs
    for orf in candidate_orfs:
        orf.logger = transcript.logger

    candidate_orfs = find_overlapping_cds(transcript, candidate_orfs)
    transcript.logger.debug("Retained %d candidate ORFS",
                            len(candidate_orfs))

    if candidate_orfs is None or len(candidate_orfs) == 0:
        transcript.logger.debug("No ORF for %s; %s", transcript.id, candidate_orfs)
        return

    transcript.combined_utr = []
    transcript.combined_cds = []
    transcript.internal_orfs = []
    transcript.finalized = False
    # Token to be set to False after the first CDS is exhausted
    primary_orf = True
    primary_strand = None
    # This will keep in memory the original BED12 objects
    transcript.loaded_bed12 = []

    primary_phase = None
    for orf in candidate_orfs:
        # Minimal check
        transcript.logger.debug("ORF for %s: start %s, end %s (%s), phase %s",
                                transcript.id, orf.thick_start, orf.thick_end, orf.end, orf.phase)
        if primary_orf is False and orf.strand != primary_strand:
            continue

        check_sanity = (orf.thick_start >= 1 and orf.thick_end <= transcript.cdna_length)
        if len(orf) != transcript.cdna_length or not check_sanity:
            message = "Wrong ORF for {0}: ".format(orf.id)
            message += "cDNA length: {0}; ".format(transcript.cdna_length)
            message += "orf length: {0}; ".format(len(orf))
            message += "CDS: {0}-{1}".format(orf.thick_start, orf.thick_end)
            transcript.logger.warning(message)
            continue

        if transcript.strand is None:
            transcript.strand = orf.strand

        transcript.loaded_bed12.append(orf)
        cds_exons = __create_internal_orf(transcript, orf)
        cds_exons = sorted(cds_exons,
                           key=operator.itemgetter(1))

        transcript.internal_orfs.append(cds_exons)

        if primary_orf is True:
            transcript.logger.debug("%s has start codon: %s, %s",
                                    transcript.id,
                                    orf.has_start_codon,
                                    orf.has_stop_codon)
            (transcript.has_start_codon, transcript.has_stop_codon) = (orf.has_start_codon,
                                                                       orf.has_stop_codon)
            primary_phase = orf.phase
            primary_orf = False
            primary_strand = orf.strand
            transcript.selected_internal_orf_index = transcript.internal_orfs.index(cds_exons)
        # transcript.phases.append(cds_exons[0][0], 0)

    # Now verify the loaded content
    check_loaded_orfs(transcript, primary_phase=primary_phase)


def check_loaded_orfs(transcript, primary_phase=0):

    """
    This function verifies the ORF status after
    loading from an external data structure/database.

    :param transcript: the transcript instance
    :type transcript: Mikado.loci_objects.transcript.Transcript

    :param primary_phase: the starting phase of the primary ORF.
    :type primary_phase: int

    :return:
    """

    if len(transcript.internal_orf_lengths) == 0:
        transcript.logger.warning("No candidate ORF retained for %s",
                                  transcript.id)

    if len(transcript.internal_orfs) == 1:
        transcript.logger.debug("Found 1 ORF for %s", transcript.id)
        transcript.combined_cds = sorted(
            [a[1] for a in iter(_ for _ in transcript.internal_orfs[0]
                                if _[0] == "CDS")],
            key=operator.itemgetter(0, 1))
        transcript.combined_utr = sorted(
            [a[1] for a in iter(_ for _ in transcript.internal_orfs[0]
                                if _[0] == "UTR")],
            key=operator.itemgetter(0, 1)
        )

    elif len(transcript.internal_orfs) > 1:
        transcript.logger.debug("Found %d ORFs for %s",
                                len(transcript.internal_orfs),
                                transcript.id)
        cds_spans = []
        candidates = []
        for internal_cds in transcript.internal_orfs:
            candidates.extend(
                [a[1] for a in iter(tup for tup in internal_cds
                                    if tup[0] == "CDS")])

        for comm in transcript.find_communities(candidates):
            span = tuple([min(t[0] for t in comm), max(t[1] for t in comm)])
            cds_spans.append(span)

        transcript.combined_cds = sorted(cds_spans, key=operator.itemgetter(0, 1))

        # This method is probably OBSCENELY inefficient,
        # but I cannot think of a better one for the moment.

        # curr_utr_segment = None  # temporary token with the UTR positions

        # Calculate the positions inside the UTR
        utr_pos = sorted(list(set.difference(
            set.union(*[set(range(exon[0], exon[1] + 1)) for exon in transcript.exons]),
            set.union(*[set(range(cds[0], cds[1] + 1)) for cds in transcript.combined_cds])
        )))

        # This code snippet using groupby is massively more efficient
        # than the previous naive implementation
        getter = operator.itemgetter(1)
        for _, group in groupby(enumerate(utr_pos), lambda items: items[0] - items[1]):
            group = [getter(element) for element in group]
            # group = list(map(operator.itemgetter(1), group))
            if len(group) > 1:
                transcript.combined_utr.append(tuple([group[0], group[-1]]))
            else:
                transcript.combined_utr.append(tuple([group[0], group[0]]))

        # Check everything is alright
        equality = (transcript.cdna_length ==
                    transcript.combined_cds_length + transcript.combined_utr_length)

        assert equality, (transcript.cdna_length, transcript.combined_cds, transcript.combined_utr)

    if transcript.internal_orfs:
        transcript.feature = "mRNA"

    transcript.phases = dict()
    transcript.__first_phase = primary_phase
    transcript._trust_orf = False
    transcript.finalize()


def __load_blast(transcript, data_dict=None, reverse=False):

    """This method looks into the DB for hits corresponding to the desired requirements.
    Hits will be loaded into the "blast_hits" list;
    we will not store the SQLAlchemy query object,
    but rather its representation as a dictionary
    (using the Hit.as_dict() method).

    :param transcript: the Transcript instance
    :type transcript: Mikado.loci_objects.transcript.Transcript
    """

    # if self.query_id is None:
    #     return

    max_target_seqs = transcript.configuration.pick.chimera_split.blast_params.max_target_seqs
    maximum_evalue = transcript.configuration.pick.chimera_split.blast_params.evalue

    if data_dict is None:
        blast_hits_query = [_.as_dict() for _ in transcript.blast_baked(transcript.session).params(
            query=transcript.id,
            evalue=maximum_evalue)]
    else:
        blast_hits_query = data_dict.get("hits", dict()).get(transcript.id, [])

    transcript.logger.debug("Starting to load BLAST data for %s",
                              transcript.id)
    # variables which take care of defining
    # the maximum evalue and target seqs
    # We do not trust the limit in the sqlite because
    # we might lose legitimate hits with the same evalue as the
    # best ones. So we collapse all hits with the same evalue.
    previous_evalue = -1
    counter = 0
    for hit in blast_hits_query:

        if counter > max_target_seqs and previous_evalue < hit["evalue"]:
            break
        elif previous_evalue < hit["evalue"]:
            previous_evalue = hit["evalue"]

        query_frames = [_["query_frame"] for _ in hit["hsps"]]
        transcript.logger.debug("Query frames for %s: %s", transcript.id, query_frames)
        if reverse is True:
            query_frames = [_ * -1 for _ in query_frames]
            transcript.logger.debug("Query frames for %s after reversal: %s", transcript.id, query_frames)

        if any(_ < 0 for _ in query_frames):
            transcript.logger.debug("Hit %s skipped for %s as it is on opposite strand",
                                    hit["target"], transcript.id)
            continue

        counter += 1

        transcript.blast_hits.append(hit)

    transcript.logger.debug("Loaded %d BLAST hits for %s",
                            counter, transcript.id)


def _connect_to_db(transcript):

    """This method will connect to the database using the information
    contained in the JSON configuration.
    :param transcript: the Transcript instance
    :type transcript: Mikado.loci_objects.transcript.Transcript

    """

    transcript.engine = dbutils.connect(
        transcript.configuration, transcript.logger)

    transcript.sessionmaker = sessionmaker()
    transcript.sessionmaker.configure(bind=transcript.engine)
    transcript.session = transcript.sessionmaker()


def load_information_from_db(transcript, configuration, introns=None, data_dict=None):
    """This method will invoke the check for:

    :param transcript: the Transcript instance
    :type transcript: Mikado.loci_objects.transcript.Transcript

    :param configuration: Necessary configuration file
    :type json_conf: (MikadoConfiguration|DaijinConfiguration)

    :param introns: the verified introns in the Locus
    :type introns: None,set

    :param data_dict: a dictionary containing the information directly
    :type data_dict: dict

    Verified introns can be provided from outside using the keyword.
    Otherwise, they will be extracted from the database directly.
    """

    transcript.logger.debug("Loading {0}".format(transcript.id))
    transcript.configuration = configuration

    __load_verified_introns(transcript, verified_introns=introns)
    retrieve_from_dict(transcript, data_dict)
    # Finally load introns, separately
    transcript.logger.debug("Loaded data for %s", transcript.id)


def retrieve_from_dict(transcript, data_dict: dict):
    """
    Method to retrieve transcript data directly from a dictionary.

    :param transcript: the Transcript instance
    :type transcript: Mikado.loci_objects.transcript.Transcript

    :param data_dict: the dictionary with loaded data from DB
    :type data_dict: dict
    """

    transcript.logger.debug(
        "Retrieving information from DB dictionary for %s",
        transcript.id)
    # Intron data
    transcript.logger.debug("Checking introns for %s, candidates %s",
                            transcript.id,
                            sorted(transcript.introns))

    # ORF data
    trust_strand = transcript.configuration.pick.orf_loading.strand_specific
    min_cds_len = transcript.configuration.pick.orf_loading.minimal_orf_length

    transcript.logger.debug("Retrieving ORF information from DB dictionary for %s",
                            transcript.id)

    if transcript.id in data_dict.get("external", dict()):
        ext_score = data_dict["external"][transcript.id]
        transcript.external_scores.update(ext_score)
        assert set(transcript.external_scores.keys()).issubset(set(ext_score.keys()))
    elif data_dict.get("external", dict()):
        raise KeyError("{} not found in external scores".format(transcript.id))

    if transcript.is_reference is False:
        if transcript.id in data_dict.get("orfs", dict()):
            candidate_orfs = list(orf for orf in data_dict["orfs"][transcript.id] if
                                  orf.cds_len >= min_cds_len)
        else:
            candidate_orfs = []

        # They must already be as ORFs
        if (transcript.monoexonic is False) or (transcript.monoexonic is True and trust_strand is True and
                                                transcript.strand is not None):
            # Remove negative strand ORFs for multiexonic transcripts,
            # or monoexonic strand-specific transcripts
            candidate_orfs = list(orf for orf in candidate_orfs if orf.strand != "-")

        old_strand = transcript.strand
        load_orfs(transcript, candidate_orfs)

        if transcript.monoexonic is False:
            is_reversed = False
        elif old_strand != transcript.strand and ((old_strand is None and transcript.strand == "-")
                                                  or (old_strand is not None)):
            is_reversed = True
        else:
            is_reversed = False
    else:
        transcript.logger.debug("Skipping ORF loading for reference %s", transcript.id)
        is_reversed = False

    transcript.logger.debug("Retrieving BLAST hits for %s",
                            transcript.id)

    __load_blast(transcript, data_dict=data_dict, reverse=is_reversed)

    transcript.logger.debug("Retrieved information from DB dictionary for %s",
                            transcript.id)


def find_overlapping_cds(transcript, candidates: list) -> List[BED12]:
    """
    Method to identify overlapping ORFs using a clique-algorithm. If any is found, ORFs are discarded until
    all ORFs are isolated from each other.
    This method will also discard ORFs that are marked as invalid by the BED12 class and/or do not meet the minimum
    requirements set up in the configuration.

    In case of overlapping ORFs, Mikado will prefere

    :param transcript: the Transcript instance
    :type transcript: Mikado.transcripts.Transcript

    :param candidates: candidate ORFs to analyse
    :type candidates: list[BED12]

    """

    # If we are looking at a multiexonic transcript
    if not (transcript.monoexonic is True and transcript.strand is None):
        candidates = list(corf for corf in candidates if corf.strand == "+")

    # Prepare the minimal secondary length parameter
    minimal_secondary_orf_length = transcript.configuration.pick.orf_loading.minimal_secondary_orf_length
    transcript.logger.debug("Minimal orf loading: %d", minimal_secondary_orf_length)

    transcript.logger.debug("{0} input ORFs for {1}".format(len(candidates), transcript.id))
    if any(corf.transcriptomic is False for corf in candidates):
        transcript.logger.debug("%d non-transcriptomic ORFs in the candidates",
                                len([corf.transcriptomic is False for corf in candidates]))
    if any(corf.invalid is True for corf in candidates):
        for corf in candidates:
            if not corf.invalid is True:
                continue
            else:
                transcript.logger.debug("Invalid ORF, reason: %s", corf.invalid_reason)

    candidates = list(corf for corf in candidates if (
        corf.invalid is False and corf.transcriptomic is True))

    ids = set(_.name for _ in candidates)
    if len(ids) < len(candidates):
        transcript.logger.debug("Colliding IDs found for the ORFs. Redefining them.")
        for pos in range(1, len(candidates) + 1):
            candidates[pos - 1].name = "{transcript.id}.orf{pos}".format(**locals())

    transcript.logger.debug("{0} filtered ORFs for {1}".format(len(candidates), transcript.id))
    if len(candidates) == 0:
        return []

    orf_dictionary = dict((x.name, x) for x in candidates)

    # First define the graph
    graph = define_graph(orf_dictionary, inters=transcript.is_overlapping_cds)
    candidate_orfs = find_candidate_orfs(transcript, graph, orf_dictionary)

    transcript.logger.debug("{0} candidate retained ORFs for {1}: {2}".format(
        len(candidate_orfs),
        transcript.id,
        [x.name for x in candidate_orfs]))
    final_orfs = [candidate_orfs[0]]
    if len(candidate_orfs) > 1:
        others = []
        for corf in candidate_orfs[1:]:
            passed = corf.cds_len >= minimal_secondary_orf_length
            transcript.logger.debug("ORF %s for %s %s minimal length of %s (its length: %s)",
                                    corf.name, transcript.id, "has" if passed else "does not have",
                                    minimal_secondary_orf_length, corf.cds_len)
            if passed:
                others.append(corf)

        final_orfs.extend(others)

    transcript.logger.debug("Retained %d ORFs for %s: %s",
                            len(final_orfs),
                            transcript.id,
                            [orf.name for orf in final_orfs])
    return final_orfs


def __create_internal_orf(transcript, orf):

    """
    Private method that calculates the assignment of the exons given the
    coordinates of the transcriptomic ORF.

    :param transcript: the Transcript instance
    :type transcript: Mikado.loci_objects.transcript.Transcript

    :param orf: candidate ORF to transform into an internal ORF
    :type orf: Mikado.serializers.orf.Orf

    """

    cds_exons = []
    current_start, current_end = 0, 0

    transcript.logger.debug("Initial phase for %s: %s", transcript.id, orf.phase)
    if orf.strand == "-":
        phase = orf.phase
        # We might decide to remove this check
        assert transcript.monoexonic is True
        current_end = transcript.start + (orf.thick_start - 1)
        current_start = current_end + orf.cds_len - 1

        assert transcript.start <= current_end <= transcript.end
        assert transcript.start <= current_end < current_start <= transcript.end, (
            transcript.start, current_end, current_start, transcript.end
        )
        cds_exons.append(("exon", transcript.exons[0]))
        if current_end > transcript.start:
            cds_exons.append(("UTR", tuple([transcript.start, current_end - 1])))
        cds_exons.append(("CDS", tuple([current_end, current_start]), phase))
        if current_start < transcript.end:
            cds_exons.append(("UTR", tuple([current_start + 1, transcript.end])))
        transcript.strand = "-"
    else:
        previous = -orf.phase
        for exon in sorted(transcript.exons, key=operator.itemgetter(0, 1),
                           reverse=(transcript.strand == "-")):
            cds_exons.append(("exon", tuple([exon[0], exon[1]])))
            current_start += 1
            current_end += exon[1] - exon[0] + 1
            # Whole UTR
            if current_end < orf.thick_start or current_start > orf.thick_end:
                cds_exons.append(("UTR", tuple([exon[0], exon[1]])))
            else:
                if transcript.strand == "+":
                    c_start = exon[0] + max(0, orf.thick_start - current_start)
                    c_end = exon[1] - max(0, current_end - orf.thick_end)
                else:
                    c_start = exon[0] + max(0, current_end - orf.thick_end)
                    c_end = exon[1] - max(0, orf.thick_start - current_start)

                if c_start > exon[0]:
                    u_end = c_start - 1
                    cds_exons.append(("UTR", tuple([exon[0], u_end])))
                if c_start <= c_end:
                    phase = (3 - (previous % 3)) % 3
                    previous += c_end - c_start + 1
                    cds_exons.append(("CDS", tuple([c_start, c_end]), phase))
                if c_end < exon[1]:
                    cds_exons.append(("UTR", tuple([c_end + 1, exon[1]])))
            current_start = current_end
        if orf.phase != 0:
            transcript.logger.debug("Non-0 phase (%d) for %s [orf: %s]",
                                    orf.phase,
                                    transcript.id,
                                    cds_exons)

    return cds_exons


def __load_verified_introns(transcript, verified_introns=None):

    """This method will load verified junctions from outside.
    Verified introns can either be given as a set or alternatively they will be retrieved by opening
    a connection to the Mikado database, using the information provided in the configuration.

    :param transcript: the Transcript instance
    :type transcript: Mikado.loci_objects.transcript.Transcript

    :param verified_introns: verified introns
    :type verified_introns: (set | None)
    """

    transcript.logger.debug("Checking introns; candidates %s", transcript.introns)
    if verified_introns is None:
        transcript.logger.debug("Checking introns using the database for %s",
                                transcript.id)

        for intron in transcript.introns:
            # Disable checks as the hybridproperties confuse both pycharm and pylint
            # noinspection PyCallByClass,PyTypeChecker
            # pylint: disable=no-value-for-parameter
            for ver_intron in transcript.session.query(Junction).filter(and_(
                    Junction.chrom == transcript.chrom,
                    intron[0] == Junction.junction_start,
                    intron[1] == Junction.junction_end)):
                if ver_intron.strand in (transcript.strand, None):
                    transcript.logger.debug("Verified intron %s%s:%d-%d for %s",
                                            transcript.chrom, ver_intron.strand,
                                            intron[0], intron[1], transcript.id)
                    transcript.verified_introns.add(intron)

    else:
        transcript.logger.debug("Checking introns using data structure for %s; introns: %s",
                                transcript.id, verified_introns)
        for intron in transcript.introns:
            transcript.logger.debug("Checking intron %s%s:%d-%d for %s",
                                    transcript.chrom, transcript.strand,
                                    intron[0], intron[1], transcript.id)
            if (intron[0], intron[1], transcript.strand) in verified_introns:
                transcript.logger.debug("Verified intron %s%s:%d-%d for %s",
                                        transcript.chrom, transcript.strand,
                                        intron[0], intron[1], transcript.id)
                transcript.verified_introns.add(intron)
            elif (intron[0], intron[1], None) in verified_introns:
                transcript.logger.debug("Verified intron %s%s:%d-%d for %s",
                                        transcript.chrom, None,
                                        intron[0], intron[1], transcript.id)
                transcript.verified_introns.add(intron)

    transcript.logger.debug("Found these introns for %s: %s",
                            transcript.id, transcript.verified_introns)
    return


def orf_sorter(orf):
    """Sorting function for the ORFs."

    :param orf: an ORF to sort
    :type orf: Mikado.serializers.orf.Orf
    """
    return (orf.cds_len,
            (orf.has_start_codon and orf.has_stop_codon),
            (orf.has_start_codon or orf.has_stop_codon),
            )


def find_candidate_orfs(transcript, graph, orf_dictionary) -> list:

    """
    Function that returns the best non-overlapping ORFs

    :param transcript: the Transcript instance
    :type transcript: Mikado.loci_objects.transcript.Transcript

    :param graph: The NetworkX graph to be analysed
    :type graph: networkx.Graph

    :param orf_dictionary: a dictionary which contains the orf indexed by name
    :type orf_dictionary: dict

    :return:
    """

    candidate_orfs = []

    while len(graph) > 0:
        cliques = find_cliques(graph, logger=transcript.logger)
        communities = find_communities(graph, logger=transcript.logger)
        clique_str = []
        for clique in cliques:
            clique_str.append(str([(orf_dictionary[x].thick_start,
                                    orf_dictionary[x].thick_end) for x in clique]))
        comm_str = []
        for comm in communities:
            comm_str.append(str([(orf_dictionary[x].thick_start,
                                  orf_dictionary[x].thick_end) for x in comm]))
        transcript.logger.debug("{0} communities for {1}:\n\t{2}".format(
            len(communities),
            transcript.id,
            "\n\t".join(comm_str)))
        transcript.logger.debug("{0} cliques for {1}:\n\t{2}".format(
            len(cliques),
            transcript.id,
            "\n\t".join(clique_str)))

        to_remove = set()
        for comm in communities:
            comm = [orf_dictionary[x] for x in comm]
            best_orf = sorted(comm, key=orf_sorter, reverse=True)[0]
            candidate_orfs.append(best_orf)
            for clique in iter(cl for cl in cliques if best_orf.name in cl):
                to_remove.update(clique)
        graph.remove_nodes_from(to_remove)

    candidate_orfs = sorted(candidate_orfs, key=orf_sorter, reverse=True)
    return candidate_orfs



from ..parsers.bed12 import BED12, _translate_str, standard, get_tables
from ..transcripts import Transcript
import unittest
from Bio.Data import CodonTable


class TestTranslate(unittest.TestCase):

    """
    >>>
    >>> table = CodonTable.ambiguous_dna_by_id[1]
    >>> _translate_str("AAA", table)
    'K'
    >>> _translate_str("TAR", table)
    '*'
    >>> _translate_str("TAN", table)
    'X'
    >>> _translate_str("TAN", table, pos_stop="@")
    '@'
    >>> _translate_str("TA?", table)

    >>> _translate_str("ATGCCCTAG", table, cds=True)
    'MP'
    >>> _translate_str("AAACCCTAG", table, cds=True)
    Traceback (most recent call last):
       ...
    Bio.Data.CodonTable.TranslationError: First codon 'AAA' is not a start codon
    >>> _translate_str("ATGCCCTAGCCCTAG", table, cds=True)
    """

    def test_from_Bio(self):

        self.assertEqual(_translate_str("TAN", standard, pos_stop="X"), "X")
        for codon, amino in standard.forward_table.forward_table.items():
            self.assertEqual(_translate_str(codon, standard), amino)

        self.assertEqual(_translate_str("TAR", standard), "*")
        self.assertEqual(_translate_str("TAR", standard, stop_symbol="?"), "?")

        self.assertEqual(_translate_str("TAN", standard, pos_stop="U"), "U")
        self.assertEqual(_translate_str("TAN", standard, pos_stop=b"U"), "U")
        with self.assertRaises(CodonTable.TranslationError):
            _translate_str("TA?", standard)

        ncbi_standard = CodonTable.ambiguous_dna_by_id[1]
        self.assertNotEqual(ncbi_standard.start_codons, standard.start_codons)
        self.assertEqual(ncbi_standard.stop_codons, standard.stop_codons)
        self.assertEqual(_translate_str("ATGCCCTAG", standard, cds=True), "MP*")
        self.assertEqual(_translate_str("ATGCCCTAG", standard, to_stop=True), "MP")
        with self.assertRaises(CodonTable.TranslationError):
            _translate_str("AAACCCTAG", standard, cds=True)
        with self.assertRaises(CodonTable.TranslationError) as exc:
            _translate_str("ATGCCCTAGCCCTAG", standard, cds=True)
        self.assertTrue(str(exc.exception).startswith("Extra in-frame stop codon found."),
                        str(exc.exception))
        with self.assertRaises(CodonTable.TranslationError) as exc:
            _translate_str("ATGCCCTAGCCCTAT", standard, cds=True)
        self.assertTrue(str(exc.exception).startswith("Extra in-frame stop codon. Sequence:"),
                        str(exc.exception))
        for invalid in (10, "AB", b"NT"):
            with self.assertRaises(ValueError):
                _translate_str("ATGCCCTAG", standard, cds=True, gap=invalid)

        with self.assertRaises(CodonTable.TranslationError):
            _translate_str("ATGC?G", standard)

    def test_invalid_pos_stop(self):
        for invalid in (10, "AB", b"NT"):
            with self.subTest(invalid=invalid), self.assertRaises(ValueError):
                _translate_str("ATGCCCTAG", standard, cds=True, pos_stop=invalid)

    def test_ncbi_standard(self):
        standard = CodonTable.ambiguous_dna_by_id[1]
        for codon, amino in standard.forward_table.forward_table.items():
            self.assertEqual(_translate_str(codon, standard), amino)

        self.assertEqual(_translate_str("TAR", standard), "*")
        self.assertEqual(_translate_str("TAR", standard, stop_symbol="?"), "?")

        self.assertEqual(_translate_str("TAN", standard, pos_stop="U"), "U")
        with self.assertRaises(CodonTable.TranslationError):
            _translate_str("TA?", standard)

        self.assertIn("CTG", standard.start_codons)
        self.assertEqual(_translate_str("ATGCCCTAG", standard, cds=True), "MP*")
        self.assertEqual(_translate_str("ATGCCCTAG", standard, to_stop=True), "MP")
        self.assertEqual(_translate_str("CTGCCCTAG", standard, cds=True), "MP*")
        self.assertEqual(_translate_str("CTGCCCTAG", standard, cds=True, to_stop=True), "MP")

    def test_ambiguous(self):
        ambigouous = None
        for key, table in CodonTable.ambiguous_dna_by_id.items():
            amb = 0 < len([c for c in table._codon_table.forward_table.keys() if c in table.stop_codons])
            if amb:
                amb = table
                break
        if ambigouous is None:
            return  # Nothing to test

    def test_set_table(self):
        b = BED12()
        for invalid in (True, list(), "Inexistent", b"Standard2"):
            with self.assertRaises(ValueError):
                b.table = invalid
        self.assertNotIn(0, CodonTable.ambiguous_dna_by_id.keys())
        for num in range(0, max(CodonTable.ambiguous_dna_by_id.keys()) + 10):
            if num in CodonTable.ambiguous_dna_by_id.keys():
                b.table = num
                self.assertEqual(b.table, CodonTable.ambiguous_dna_by_id[num])
            elif num == 0:
                b.table = num
                self.assertEqual(b.table, standard)
            else:
                with self.assertRaises(ValueError):
                    b.table = num

        for valid in list(CodonTable.ambiguous_dna_by_name.keys()):
            b.table = valid
            self.assertEqual(b.table, CodonTable.ambiguous_dna_by_name[valid])


class Bed12GenToTrans(unittest.TestCase):

    def setUp(self):
        pass

    def test_positive_mono_transfer(self):

        string_bed = "1\t10\t500\ttest\t0\t+\t300\t390\t0\t1\t490\t0"


        bed = BED12(string_bed)
        self.assertFalse(bed.invalid)
        self.assertFalse(bed.header)
        self.assertEqual(bed.thick_start, 301)
        self.assertEqual(bed.thick_end, 390)
        self.assertEqual(bed.blocks, [(11, 500)])

        seq = "AAA" * 96 + "AA"
        seq += "ATG" * 29
        seq += "TGA"
        seq += "A" * 110
        self.assertEqual(len(seq), len(bed))

        tbed = bed.to_transcriptomic(sequence=seq)
        self.assertEqual(tbed.start, 1)
        self.assertEqual(tbed.end, 500 - 11 + 1)
        self.assertEqual(tbed.thick_end - tbed.thick_start + 1, 90)
        self.assertEqual(tbed.thick_start, 291)
        self.assertEqual(tbed.thick_end, 380)
        self.assertTrue(tbed.has_start_codon)
        self.assertTrue(tbed.has_stop_codon)
        self.assertEqual(seq[tbed.thick_start - 1:tbed.thick_end], "ATG" * 29 + "TGA")

    def test_negative_mono_transfer(self):

        string_bed = "1\t10\t500\ttest\t0\t-\t300\t390\t0\t1\t490\t0"

        bed = BED12(string_bed)
        self.assertFalse(bed.invalid)
        self.assertFalse(bed.header)
        self.assertEqual(bed.thick_start, 301)
        self.assertEqual(bed.thick_end, 390)
        self.assertEqual(bed.blocks, [(11, 500)])

        seq = "A" * 110
        seq += "ATG" * 29
        seq += "TGA"
        seq += "A" * (490 - 110 - 90)
        self.assertEqual(len(seq), len(bed))

        tbed = bed.to_transcriptomic(sequence=seq)
        self.assertEqual(tbed.start, 1)
        self.assertEqual(tbed.end, 500 - 11 + 1)

        self.assertEqual(tbed.thick_end - tbed.thick_start + 1, 90)
        self.assertEqual(tbed.thick_start, 111)
        self.assertEqual(tbed.thick_end, 200)
        self.assertTrue(tbed.has_start_codon)
        self.assertTrue(tbed.has_stop_codon)

        self.assertEqual(seq[tbed.thick_start - 1:tbed.thick_end], "ATG" * 29 + "TGA")

    def test_diexonic_pos_transfer(self):

        string_bed = "1\t10\t1000\ttest\t0\t+\t80\t920\t0\t2\t190,200\t0,790"
        bed = BED12(string_bed)
        self.assertFalse(bed.invalid or bed.header)
        self.assertEqual(bed.start, 11)
        self.assertEqual(bed.end, 1000)
        self.assertEqual(bed.blocks, [(11, 200), (801, 1000)])
        self.assertEqual(bed.thick_start, 81)
        self.assertEqual(bed.thick_end, 920)

        seq = "A" * 70
        seq += "ATG" * 79
        seq += "TAA" * 1
        seq += "A" * 80
        self.assertEqual(len(seq), bed.block_sizes.sum())
        tbed = bed.to_transcriptomic(sequence=seq)
        self.assertEqual(tbed.start, 1)
        self.assertEqual(tbed.end, 390)
        self.assertEqual(tbed.thick_end - tbed.thick_start + 1, 240)
        self.assertEqual(tbed.thick_start, 71)
        self.assertEqual(tbed.thick_end, 310)
        self.assertTrue(tbed.has_start_codon)
        self.assertTrue(tbed.has_stop_codon)

        self.assertEqual(seq[tbed.thick_start - 1:tbed.thick_end], "ATG" * 79 + "TAA")

    def test_diexonic_neg_transfer(self):

        string_bed = "1\t10\t1000\ttest\t0\t-\t80\t920\t0\t2\t190,200\t0,790"
        bed = BED12(string_bed)
        self.assertFalse(bed.invalid or bed.header)
        self.assertEqual(bed.start, 11)
        self.assertEqual(bed.end, 1000)
        self.assertEqual(bed.blocks, [(11, 200), (801, 1000)])
        self.assertEqual(bed.thick_start, 81)
        self.assertEqual(bed.thick_end, 920)

        seq = "A" * 80
        seq += "ATG" * 79
        seq += "TAA" * 1
        seq += "A" * 70
        self.assertEqual(len(seq), bed.block_sizes.sum())
        tbed = bed.to_transcriptomic(sequence=seq)
        self.assertEqual(tbed.start, 1)
        self.assertEqual(tbed.end, 390)
        self.assertEqual(tbed.thick_end - tbed.thick_start + 1, 240)
        self.assertEqual(tbed.thick_start, 81)
        self.assertEqual(tbed.thick_end, 320)
        self.assertTrue(tbed.has_start_codon)
        self.assertTrue(tbed.has_stop_codon)

        self.assertEqual(seq[tbed.thick_start - 1:tbed.thick_end], "ATG" * 79 + "TAA")

    def test_wheat_1(self):

        string_bed = "chr7A\t207087445\t207089574\tTraesCS7A01G235400.1\t0\t-\t207087615\t207088433\t0\t3\t457,393,30\t0,603,2099"
        string_seq = """CGCGTCGGTGCATCCGGATACGTCGCCTGGGCTACACAATGGCGCTGATCGATTGGATAG
AACTGAGTGATGATGCAGAGATTATTGAATTGAGCAGTAGCGAGGAGAATGTCGAAGAAT
ATCAGGGAGCTTCCACACAGAGCAACTCAGCTCAGCACCAAGCTACATTGCATGATGATC
AGACCATGTTTGTCACAGAAGGTGAACGAAGGGAAGAGGCTACCGAATCTGGAAATGCAG
AGGAAGCTACTGCATCCTCCTCCGTTACTGAAAAAGAGTCTGGAAATCCAGAGGAAGCTA
CTGTATCAGAGACTTGTCCTCACACTCCAACAGCCGTTGTTACTCTAAAAGAGTCCAGAA
ATCCAGATGAAGCTACTGTATCACAAACTTGTCCTCACACTACAACAGCTGTGACGTTCC
CCGGACCTGATCCCTTCTCTCCGAAGGCTCTGACTTTTCAAGCCTGTGTCGCCAAGCCGC
GGAGAGCGAAGGCGAAACGTCGCAGGAAGCTCTTCCACGCCGATACACCTTGGACTTGGA
GGAGCCCTAGGCTTGAAGACAAGCATAAAGGTCGTAGAAGGCCCGTGGAAGAGCTGGCAG
TTGAGCGCAAGGCTTCTGCCATGCCGGAAGAAGAAGAACAGAAGAACACCTCGGCTAGGA
GCCGGAGGAAATGCAAGGTGTTCGGCAAACGGTGTAACCTCGGCAGATAGCCTCCACTGA
TCCCGCAAAACTAAGGTAGTCGACTGAAGATGAAAGATGATCGTCTCGAAACTGTCTGGA
TTTTGGCGGTCCAGTTTGCCTCTCCTGTGACGCGGGAAGAATCACTAGTGTTGGTGGGGT
TTGGTGGGGTCGTGTTGCCCTTGCTGCGCTGTCCATATGA"""

        string_seq = "".join(string_seq.split("\n"))

        string_cds = """ATGGCGCTGATCGATTGGATAGAACTGAGTGATGATGCAGAGATTATTGAATTGAGCAGT
AGCGAGGAGAATGTCGAAGAATATCAGGGAGCTTCCACACAGAGCAACTCAGCTCAGCAC
CAAGCTACATTGCATGATGATCAGACCATGTTTGTCACAGAAGGTGAACGAAGGGAAGAG
GCTACCGAATCTGGAAATGCAGAGGAAGCTACTGCATCCTCCTCCGTTACTGAAAAAGAG
TCTGGAAATCCAGAGGAAGCTACTGTATCAGAGACTTGTCCTCACACTCCAACAGCCGTT
GTTACTCTAAAAGAGTCCAGAAATCCAGATGAAGCTACTGTATCACAAACTTGTCCTCAC
ACTACAACAGCTGTGACGTTCCCCGGACCTGATCCCTTCTCTCCGAAGGCTCTGACTTTT
CAAGCCTGTGTCGCCAAGCCGCGGAGAGCGAAGGCGAAACGTCGCAGGAAGCTCTTCCAC
GCCGATACACCTTGGACTTGGAGGAGCCCTAGGCTTGAAGACAAGCATAAAGGTCGTAGA
AGGCCCGTGGAAGAGCTGGCAGTTGAGCGCAAGGCTTCTGCCATGCCGGAAGAAGAAGAA
CAGAAGAACACCTCGGCTAGGAGCCGGAGGAAATGCAAGGTGTTCGGCAAACGGTGTAAC
CTCGGCAGATAG"""
        string_cds = "".join(string_cds.split("\n"))

        bed = BED12(string_bed)
        self.assertFalse(bed.invalid or bed.header)
        self.assertEqual(bed.start, 207087445+1)
        self.assertEqual(bed.end, 207089574)
        self.assertEqual(bed.strand, "-")
        self.assertEqual(bed.blocks, [(207087446,207087902), (207088049,207088441), (207089545,207089574) ])
        self.assertEqual(bed.thick_start, 207087616)
        self.assertEqual(bed.thick_end, 207088433)

        self.assertEqual(len(string_seq), bed.block_sizes.sum())
        tbed = bed.to_transcriptomic(sequence=string_seq)
        self.assertEqual(tbed.thick_end - tbed.thick_start + 1, 672)
        self.assertEqual(tbed.thick_start, string_seq.index("ATGGCGCTGATCGATTGGA") + 1)
        self.assertEqual(tbed.thick_start, 39)
        self.assertEqual(tbed.thick_end, string_seq.index("CTCGGCAGATAG") + len("CTCGGCAGATAG"))
        self.assertEqual(string_cds, string_seq[tbed.thick_start - 1:tbed.thick_end])

    def test_mono_pos_bed_with_phase(self):

        string = "1\t10\t101\tID=test;phase=2;coding=True\t0\t+\t10\t101\t0\t1\t91\t0"

        seq = "A" + "CGG" * 29 + "TAA"

        bed = BED12(string, transcriptomic=False)
        self.assertFalse(bed.header)
        self.assertEqual(bed.thick_start, 11)
        self.assertFalse(bed.invalid)
        self.assertEqual(bed.name, "test")
        self.assertEqual(bed.phase, 2)

        tbed = bed.to_transcriptomic(sequence=seq)

        self.assertTrue(tbed.has_stop_codon)
        self.assertFalse(tbed.has_start_codon)
        self.assertEqual(tbed.thick_start, 1)
        self.assertEqual(tbed.thick_end, 91)
        self.assertFalse(tbed.invalid)
        self.assertEqual(tbed.phase, 2)
        self.assertTrue(tbed.transcriptomic)

    def test_mono_neg_bed_with_phase(self):

        string = "1\t10\t101\tID=test;phase=2;coding=True\t0\t-\t10\t101\t0\t1\t91\t0"

        seq = "A" + "CGG" * 29 + "TAA"

        bed = BED12(string, transcriptomic=False)
        self.assertFalse(bed.header)
        self.assertEqual(bed.thick_start, 11)
        self.assertFalse(bed.invalid)
        self.assertEqual(bed.name, "test")
        self.assertEqual(bed.phase, 2)

        tbed = bed.to_transcriptomic(sequence=seq)

        self.assertTrue(tbed.has_stop_codon)
        self.assertFalse(tbed.has_start_codon)
        self.assertEqual(tbed.thick_start, 1)
        self.assertEqual(tbed.thick_end, 91)
        self.assertFalse(tbed.invalid)
        self.assertEqual(tbed.phase, 2)
        self.assertTrue(tbed.transcriptomic)

    def test_diex_pos_bed_with_phase_one(self):

        string = "1\t10\t111\tID=test;phase=1;coding=True\t0\t+\t10\t101\t0\t1\t101\t0"

        seq = "A" + "CGG" * 29 + "TAA" + "A" * 10

        bed = BED12(string, transcriptomic=False)
        self.assertFalse(bed.header)
        self.assertEqual(bed.thick_start, 11)
        self.assertFalse(bed.invalid)
        self.assertEqual(bed.name, "test")
        self.assertEqual(bed.phase, 1)

        tbed = bed.to_transcriptomic(sequence=seq)

        self.assertFalse(tbed.has_start_codon)
        self.assertEqual(tbed.thick_start, 1)
        self.assertEqual(tbed.thick_end, 91)
        self.assertFalse(tbed.invalid, tbed.invalid_reason)
        self.assertEqual(tbed.phase, 1)
        self.assertTrue(tbed.transcriptomic)
        self.assertTrue(tbed.has_stop_codon)

    def test_diex_neg_bed_with_phase_one(self):

        string = "1\t10\t300\tID=test;phase=1;coding=True\t0\t-\t70\t300\t0\t2\t90,100\t0,190"

        seq = "A" + "CGG" * 42 + "TAA" + "A" * 60

        bed = BED12(string, transcriptomic=False)
        self.assertFalse(bed.header)
        self.assertEqual(bed.thick_start, 71)
        self.assertFalse(bed.invalid)
        self.assertEqual(bed.name, "test")
        self.assertEqual(bed.phase, 1)

        tbed = bed.to_transcriptomic(sequence=seq)

        self.assertFalse(tbed.has_start_codon)
        self.assertEqual(tbed.thick_start, 1)
        self.assertEqual(tbed.thick_end, 130)
        self.assertEqual(tbed.phase, 1)
        self.assertFalse(tbed.invalid, tbed.invalid_reason)
        self.assertTrue(tbed.transcriptomic)
        self.assertTrue(tbed.has_stop_codon)

    def test_diex_pos_bed_with_phase_two(self):

        string = "1\t9\t111\tID=test;phase=2;coding=True\t0\t+\t9\t101\t0\t1\t102\t0"

        seq = "GA" + "CGG" * 29 + "TAA" + "A" * 10

        bed = BED12(string, transcriptomic=False)
        self.assertFalse(bed.header)
        self.assertEqual(bed.thick_start, 10)
        self.assertFalse(bed.invalid)
        self.assertEqual(bed.name, "test")
        self.assertEqual(bed.phase, 2)

        tbed = bed.to_transcriptomic(sequence=seq)

        self.assertFalse(tbed.has_start_codon)
        self.assertEqual(tbed.thick_start, 1)
        self.assertEqual(tbed.thick_end, 92)
        self.assertFalse(tbed.invalid, tbed.invalid_reason)
        self.assertEqual(tbed.phase, 2)
        self.assertTrue(tbed.transcriptomic)
        self.assertTrue(tbed.has_stop_codon)

    def test_diex_neg_bed_with_phase_two(self):

        string = "1\t10\t301\tID=test;phase=2;coding=True\t0\t-\t70\t301\t0\t2\t90,101\t0,190"

        seq = "GA" + "CGG" * 42 + "TAA" + "A" * 60

        bed = BED12(string, transcriptomic=False)
        self.assertFalse(bed.header)
        self.assertEqual(bed.thick_start, 71)
        self.assertFalse(bed.invalid)
        self.assertEqual(bed.name, "test")
        self.assertEqual(bed.phase, 2)

        tbed = bed.to_transcriptomic(sequence=seq)

        self.assertFalse(tbed.has_start_codon)
        self.assertEqual(tbed.thick_start, 1)
        self.assertEqual(tbed.thick_end, 131)
        self.assertEqual(tbed.phase, 2)
        self.assertFalse(tbed.invalid, tbed.invalid_reason)
        self.assertTrue(tbed.transcriptomic)
        self.assertTrue(tbed.has_stop_codon)

    def test_tran_to_bed12_neg(self):

        for end, phase in [(299, 0), (300, 1), (301, 2)]:
            with self.subTest():
                t = Transcript()
                t.chrom = "1"
                t.start, t.end, t.strand, t.id = 11, end, "-", "test"

                t.add_exons([(11, 100), (201, end)])
                t.add_exons([(71, 100), (201, end)], features="CDS", phases=[0, phase])
                t.finalize()
                r = t.as_bed12()
                self.assertEqual(r.name, t.id)
                self.assertEqual(r.phase, phase, (end, phase))
                self.assertEqual(r.thick_end, end)
                self.assertFalse(r.invalid)

    def test_touching_exons(self):

        bed12line = ["chr1", 172601, 175626, "ID=foo.1", 100, "-", 172601, 175626, "0,0,0", 3, "199,1281,861,",
                     "0,199,2164"]
        bed = BED12(bed12line, transcriptomic=False)
        self.assertFalse(bed.invalid, bed.invalid_reason)
        t = Transcript(bed)
        t.finalize()
        self.assertEqual(t.exon_num, 2)  # The touching exons should have been merged
        self.assertEqual(t.exons, [(172602, 174081), (174766, 175626)], t.exons)


if __name__ == "__main__":
    unittest.main()


import unittest
from .. import loci, transcripts
from .._transcripts.scoring_configuration import Requirements, MinMaxScore, TargetScore
from ..configuration import MikadoConfiguration
from ..parsers.bed12 import BED12
from ..transcripts import Transcript
from ..utilities.log_utils import create_default_logger


class ScoreTester(unittest.TestCase):

    def setUp(self):

        b1 = BED12("1\t100\t900\tID=t1;coding=True\t10\t+\t200\t800\t.\t1\t800\t0", coding=True, transcriptomic=False)

        self.t1 = Transcript(b1)
        self.t1.finalize()
        self.assertTrue(self.t1.is_coding)

        b2 = BED12("1\t100\t1100\tID=t2;coding=True\t20\t+\t200\t900\t.\t2\t300,300\t0,700", coding=True)
        self.t2 = Transcript(b2)
        self.t2.finalize()
        self.assertTrue(self.t2.is_coding)

        # 200, 400  - 200
        # 650, 850  - 300
        # 1200, 1360    - 160

        b3 = BED12("1\t100\t1500\tID=t3;coding=True\t30\t+\t200\t1360\t.\t3\t300,300,300\t0,550,1100", coding=True)
        self.assertFalse(b3.header)
        self.assertEqual(b3.thick_start, 201)
        self.assertFalse(b3.invalid, b3.invalid_reason)
        self.assertEqual(b3.block_count, 3)
        self.assertTrue((b3.block_sizes == [300, 300, 300]).all())
        self.t3 = Transcript(b3)
        self.t3.finalize()
        self.assertTrue(self.t3.is_coding)
        self.configuration = MikadoConfiguration()
        reqs = {"expression": ["cdna_length"],
                "parameters": {"cdna_length": {"operator": "gt", "value": 0}}
                }

        # self.configuration.requirements = reqs["requirements"]
        self.configuration.scoring.requirements = Requirements.Schema().load(reqs)
        self.configuration.check()

        self.locus = loci.Superlocus(self.t1, configuration=self.configuration)
        self.locus.add_transcript_to_locus(self.t2)
        self.locus.add_transcript_to_locus(self.t3)

    def test_exon_num_max(self):

        for multiplier in (1, 2, 3):
            with self.subTest(multiplier=multiplier):
                logger = create_default_logger("test_exon_num_max", level="WARNING")
                self.locus.configuration.scoring.scoring["exon_num"] = MinMaxScore.Schema().load(
                    {"rescaling": "max", "use_raw": False, "multiplier": multiplier})
                self.assertEqual(self.locus.configuration.scoring.scoring["exon_num"].multiplier, multiplier)
                self.assertIn("t3", self.locus.transcripts)
                self.locus.logger = logger
                self.locus.filter_and_calculate_scores()
                self.assertEqual(self.locus.scores["t1"]["exon_num"], 0)
                self.assertIn("t3", self.locus.transcripts)
                self.assertIn("t3", self.locus.scores, self.locus.scores)
                self.assertEqual(self.locus.scores["t3"]["exon_num"], multiplier,
                                 self.locus.configuration.scoring)
                self.assertEqual(self.locus.scores["t2"]["exon_num"], 0.5 * multiplier, self.locus.scores)
                self.locus.scores_calculated = False

    def test_exon_min(self):

        for multiplier in (1, 2, 3):
            with self.subTest(multiplier=multiplier):
                logger = create_default_logger("test_exon_num_max", level="WARNING")
                self.locus.configuration.scoring.scoring["exon_num"] = MinMaxScore.Schema().load(
                    {"rescaling": "min", "use_raw": False, "multiplier": multiplier})
                self.assertIn("t3", self.locus.transcripts)
                self.locus.logger = logger
                self.locus.filter_and_calculate_scores()
                self.assertEqual(self.locus.scores["t1"]["exon_num"], multiplier)
                self.assertIn("t3", self.locus.transcripts)
                self.assertIn("t3", self.locus.scores, self.locus.scores)
                self.assertEqual(self.locus.scores["t3"]["exon_num"], 0, self.locus.scores)
                self.assertEqual(self.locus.scores["t2"]["exon_num"], 0.5 * multiplier, self.locus.scores)
                self.locus.scores_calculated = False

    def test_exon_target(self):

        for multiplier in (1, 2, 3):
            with self.subTest(multiplier=multiplier):
                logger = create_default_logger("test_exon_num_max", level="WARNING")

                self.locus.configuration.scoring.scoring["exon_num"] = TargetScore.Schema().load(
                    {"rescaling": "target", "value": 2, "use_raw": False, "multiplier": multiplier})
                self.assertIn("t3", self.locus.transcripts)
                self.locus.logger = logger
                self.locus.filter_and_calculate_scores()
                self.assertEqual(self.locus.scores["t1"]["exon_num"], 0)
                self.assertIn("t3", self.locus.transcripts)
                self.assertIn("t3", self.locus.scores, self.locus.scores)
                self.assertEqual(self.locus.scores["t3"]["exon_num"], 0, self.locus.scores)
                self.assertEqual(self.locus.scores["t2"]["exon_num"], multiplier, self.locus.scores)
                self.locus.scores_calculated = False

    def test_exon_target_min(self):

        for multiplier in (1, 2, 3):
            with self.subTest(multiplier=multiplier):
                logger = create_default_logger("test_exon_num_max", level="WARNING")

                self.locus.configuration.scoring.scoring["exon_num"] = TargetScore.Schema().load(
                    {"rescaling": "target", "value": 1,
                     "use_raw": False, "multiplier": multiplier})
                self.assertIn("t3", self.locus.transcripts)
                self.locus.logger = logger
                self.locus.filter_and_calculate_scores()
                self.assertEqual(self.locus.scores["t1"]["exon_num"], multiplier)
                self.assertIn("t3", self.locus.transcripts)
                self.assertIn("t3", self.locus.scores, self.locus.scores)
                self.assertEqual(self.locus.scores["t3"]["exon_num"], 0, self.locus.scores)
                self.assertEqual(self.locus.scores["t2"]["exon_num"], 0.5 * multiplier, self.locus.scores)
                self.locus.scores_calculated = False

    def test_exon_max_filter(self):

        for multiplier in (1, 2, 3):
            with self.subTest(multiplier=multiplier):
                logger = create_default_logger("test_exon_num_max", level="WARNING")

                self.locus.configuration.scoring.scoring["exon_num"] = MinMaxScore.Schema().load(
                    {"rescaling": "max", "use_raw": False, "multiplier": multiplier,
                     "filter": {"operator": "gt", "value": 2}})
                self.assertIn("t3", self.locus.transcripts)
                self.locus.logger = logger
                self.locus.filter_and_calculate_scores()
                self.assertEqual(self.locus.scores["t1"]["exon_num"], 0)
                self.assertIn("t3", self.locus.transcripts)
                self.assertIn("t3", self.locus.scores, self.locus.scores)
                self.assertEqual(self.locus.scores["t3"]["exon_num"], multiplier, self.locus.scores)
                self.assertEqual(self.locus.scores["t2"]["exon_num"], 0, self.locus.scores)
                self.locus.scores_calculated = False

    def test_combined_cds_length_exon_filter(self):

        for multiplier in (1, 2, 3):
            with self.subTest(multiplier=multiplier):
                scoring = {"rescaling": "max", "use_raw": False, "multiplier": multiplier,
                           "filter": {"operator": "gt", "value": 2, "metric": "exon_num"}}

                logger = create_default_logger("test_exon_num_max", level="WARNING")

                self.locus.configuration.scoring.scoring = {"combined_cds_length": MinMaxScore.Schema().load(scoring)}
                self.locus.configuration = self.locus.configuration
                self.assertIn("t3", self.locus.transcripts)
                self.locus.logger = logger
                self.locus.filter_and_calculate_scores()
                self.assertEqual(self.locus.scores["t1"]["combined_cds_length"], 0)
                self.assertIn("t3", self.locus.transcripts)
                self.assertIn("t3", self.locus.scores, self.locus.scores)
                self.assertEqual(self.locus.scores["t3"]["combined_cds_length"], multiplier, self.locus.scores)
                self.assertEqual(self.locus.scores["t2"]["combined_cds_length"], 0, self.locus.scores)
                self.locus.scores_calculated = False

    def test_combined_cds_length_exon_filter_lt2(self):

        for multiplier in (1, 2, 3):
            with self.subTest(multiplier=multiplier):
                logger = create_default_logger("test_exon_num_max", level="WARNING")

                self.locus.configuration.scoring.scoring["combined_cds_length"] = MinMaxScore.Schema().load(
                    {"rescaling": "max", "use_raw": False, "multiplier": multiplier,
                     "filter": {"operator": "lt", "value": 3, "metric": "exon_num"}
                     }
                )
                self.assertIn("t3", self.locus.transcripts)
                self.locus.logger = logger
                self.locus.filter_and_calculate_scores()

                self.assertIn("t3", self.locus.transcripts)
                self.assertIn("t3", self.locus.scores, self.locus.scores)
                self.assertEqual(self.locus.scores["t3"]["combined_cds_length"], 0, self.locus.scores)
                self.assertEqual(self.locus.scores["t2"]["combined_cds_length"], 0, self.locus.scores)
                self.assertEqual(self.locus.scores["t1"]["combined_cds_length"], multiplier)
                self.locus.scores_calculated = False

    def test_default_scores(self):

        json_conf = MikadoConfiguration()
        locus = loci.Superlocus(self.t1, use_transcript_scores=True, configuration=json_conf)
        locus.add_transcript_to_locus(self.t2)
        locus.add_transcript_to_locus(self.t3)
        self.assertTrue(locus._use_transcript_scores)
        self.assertTrue(locus.scores_calculated)
        for transcript in (self.t1, self.t2, self.t3):
            self.assertIn(transcript.id, locus)
            self.assertEqual(locus.scores[transcript.id], transcript.score)
        locus.filter_and_calculate_scores()
        for transcript in (self.t1, self.t2, self.t3):
            self.assertIn(transcript.id, locus)
            self.assertEqual(locus.scores[transcript.id], transcript.score)


class LocusMissedTester(unittest.TestCase):

    def setUp(self):
        self.configuration = MikadoConfiguration()
        reqs = {
                "expression": ["cdna_length"],
                "parameters": {
                    "cdna_length": {
                        "operator": "gt", "value": 0,
                    "name": "cdna_length"}
                }
            }

        self.configuration.scoring.requirements = Requirements.Schema().load(reqs)
        self.configuration.pick.alternative_splicing.pad = False
        self.configuration.check()

    def test_transcript_not_missed(self):

        b1 = BED12("1\t100\t500\tID=t1;coding=True\t20\t+\t200\t459\t.\t2\t200,100,\t0,300", coding=True)
        self.assertFalse(b1.invalid, b1.invalid_reason)
        self.assertTrue(b1.coding)
        self.assertTrue(b1.is_transcript)
        self.assertEqual(b1.thick_start, 201)
        self.assertEqual(b1.thick_end, 459)
        logger = create_default_logger("test_transcript_missed", level="ERROR")
        t1 = transcripts.Transcript(b1, logger=logger)
        t1.finalize()
        self.assertEqual(sorted(t1.exons), [(101, 300), (401,500)])
        b2 = BED12("1\t100\t1500\tID=t2;coding=True\t25\t+\t200\t901\t.\t4\t200,150,400,200\t0,300,700,1200",
                   coding=True)
        self.assertFalse(b2.invalid, b2.invalid_reason)
        self.assertTrue(b2.coding)
        self.assertTrue(b2.is_transcript)
        self.assertEqual(b2.thick_start, 201)
        self.assertEqual(b2.thick_end, 901)
        # logger.setLevel("DEBUG")
        t2 = transcripts.Transcript(b2, logger=logger)
        t2.finalize()
        self.assertTrue(t2.is_coding)
        self.assertEqual(sorted(t2.exons), [(101, 300), (401, 550), (801, 1200), (1301, 1500)])
        self.assertEqual([t2.combined_cds_start, t2.combined_cds_end], [201, 901])

        b3 = BED12("1\t800\t1500\tID=t3;coding=True\t40\t+\t820\t1370\t.\t4\t50,100,100,200\t0,150,300,500",
                   coding=True)
        self.assertFalse(b3.invalid, b3.invalid_reason)
        self.assertTrue(b3.coding)
        self.assertTrue(b3.is_transcript)
        self.assertEqual(b3.thick_start, 821)
        self.assertEqual(b3.thick_end, 1370)
        # logger.setLevel("DEBUG")
        t3 = transcripts.Transcript(b3, logger=logger)
        t3.finalize()
        self.assertTrue(t3.is_coding)
        self.assertEqual(sorted(t3.exons), [(801, 850), (951, 1050), (1101, 1200), (1301, 1500)])
        self.assertEqual([t3.combined_cds_start, t3.combined_cds_end], [821, 1370])

        locus = loci.Superlocus(t1, use_transcript_scores=True, configuration=self.configuration, logger=logger)
        locus.add_transcript_to_locus(t2)
        locus.add_transcript_to_locus(t3)
        locus.define_loci()
        self.assertEqual(len(locus.loci), 2)
        primaries = set([locus.loci[_].primary_transcript_id for _ in locus.loci])
        self.assertEqual(primaries, {t3.id, t1.id})

    def test_transcript_missed(self):

        """This unit-test describes a situation which currently is pathological in Mikado (issue #131).
        Namely, given three transcripts (A, B, C), with A sharing an intron with B and having *lower* score;
        and B sharing a *splice site* with C but having lower score than the latter; Mikado will end up selecting
        B and C at the sublocus stage, then discard B - without recovering A. This will lead to potential loss of good
        loci.
        """

        b1 = BED12("1\t100\t500\tID=t1;coding=True\t20\t+\t200\t459\t.\t2\t200,100,\t0,300", coding=True)
        self.assertFalse(b1.invalid, b1.invalid_reason)
        self.assertTrue(b1.coding)
        self.assertTrue(b1.is_transcript)
        self.assertEqual(b1.thick_start, 201)
        self.assertEqual(b1.thick_end, 459)
        logger = create_default_logger("test_transcript_missed", level="ERROR")
        t1 = transcripts.Transcript(b1, logger=logger)
        t1.finalize()
        self.assertEqual(sorted(t1.exons), [(101, 300), (401, 500)])

        b2 = BED12("1\t100\t1500\tID=t2;coding=True\t25\t+\t200\t901\t.\t4\t200,150,400,170\t0,300,700,1230",
                   coding=True)
        self.assertFalse(b2.invalid, b2.invalid_reason)
        self.assertTrue(b2.coding)
        self.assertTrue(b2.is_transcript)
        self.assertEqual(b2.thick_start, 201)
        self.assertEqual(b2.thick_end, 901)
        # logger.setLevel("DEBUG")
        t2 = transcripts.Transcript(b2, logger=logger)
        t2.finalize()
        self.assertTrue(t2.is_coding)
        self.assertEqual(sorted(t2.exons), [(101, 300), (401, 550), (801, 1200), (1331, 1500)])
        self.assertEqual([t2.combined_cds_start, t2.combined_cds_end], [201, 901])

        b3 = BED12("1\t800\t1500\tID=t3;coding=True\t40\t+\t820\t1370\t.\t4\t50,100,100,200\t0,150,300,500",
                   coding=True)
        self.assertFalse(b3.invalid, b3.invalid_reason)
        self.assertTrue(b3.coding)
        self.assertTrue(b3.is_transcript)
        self.assertEqual(b3.thick_start, 821)
        self.assertEqual(b3.thick_end, 1370)
        # logger.setLevel("DEBUG")
        t3 = transcripts.Transcript(b3, logger=logger)
        t3.finalize()
        self.assertTrue(t3.is_coding)
        self.assertEqual(sorted(t3.exons), [(801, 850), (951, 1050), (1101, 1200), (1301, 1500)])
        self.assertEqual([t3.combined_cds_start, t3.combined_cds_end], [821, 1370])

        locus = loci.Superlocus(t1, use_transcript_scores=True, configuration=self.configuration, logger=logger)
        locus.add_transcript_to_locus(t2)
        locus.add_transcript_to_locus(t3)
        locus.define_loci()
        self.assertEqual(len(locus.loci), 2)
        primaries = set([locus.loci[_].primary_transcript_id for _ in locus.loci])
        self.assertEqual(primaries, {t3.id, t1.id})
        # self.assertEqual(set(locus.lost_transcripts.keys()), {t1.id})



#!/usr/bin/env python3
# coding: utf-8

""" Script to remove sequences specific of a given organism from a SwissProt file. """

import sys
import argparse
from Bio import SeqIO
import os
import re
import gzip


def main():
    """Main script function"""
    def to_file(string):
        """ Function to check the correct parser for the input file.
        :param string: filename
        """
        assert os.path.exists(string)
        if string.endswith(".gz"):
            handle = gzip.open(string)
        else:
            handle = open(string)
        return SeqIO.parse(handle, "swiss")

    parser = argparse.ArgumentParser("Script to remove sequences specific of a given organism from a SwissProt file.")
    parser.add_argument("-o", "--organism", type=str, help="Organism to be excluded", required=True)
    parser.add_argument("--format", type=str, choices=["fasta"], default="fasta",
                        help="Output format. Choices: %(choices)s. Default: %(default)s.")
    parser.add_argument("input", type=to_file)
    parser.add_argument("out", default=sys.stdout, type=argparse.FileType("w"), nargs="?")
    args = parser.parse_args()

    for record in args.input:
        if re.search(args.organism, record.annotations["organism"]) is None:
            SeqIO.write(record, args.out, args.format)
        else:
            continue


if __name__ == "__main__":
    main()



from .transcript import Transcript
import pysam
from ..exceptions import InvalidTranscript
from .transcriptchecker import TranscriptChecker


def pad_transcript(transcript: Transcript,
                   backup: Transcript,
                   start_transcript: [Transcript, bool],
                   end_transcript: [Transcript, bool],
                   fai: pysam.libcfaidx.FastaFile,
                   logger):

    """This method will enlarge the coordinates and exon structure of a transcript, given:
    :param transcript: the transcript to modify.
    :type transcript: Transcript
    :param backup: a copy of the transcript to be modified.
    :type backup: Transcript
    :param start_transcript: the template transcript for the 5' end.
    :param end_transcript: the template transcript for the 3' end.
    :param fai: the indexed genomic sequence.
    :param logger: the logger to be used in the function.
    """

    # If there is nothing to do, just get out
    assert transcript == backup
    transcript.finalize()
    if start_transcript not in (False, None):
        start_transcript.finalize()
    if end_transcript not in (False, None):
        end_transcript.finalize()

    if start_transcript in (False, None) and end_transcript in (False, None):
        logger.debug("%s does not need to be expanded, exiting", transcript.id)
        return transcript

    if transcript.strand == "-":
        start_transcript, end_transcript = end_transcript, start_transcript

    # Make a backup copy of the transcript
    # First get the ORFs
    # Remove the CDS and unfinalize
    logger.debug("Starting expansion of %s", transcript.id)
    strand = transcript.strand
    transcript.strip_cds()
    transcript.unfinalize()
    assert strand == transcript.strand

    upstream = 0
    up_exons = []
    up_remove = False
    downstream = 0
    down_exons = []
    down_remove = False
    new_exon = None
    try:
        upstream, up_exons, new_first_exon, up_remove = _enlarge_start(transcript, backup, start_transcript)
        downstream, up_exons, down_exons, new_exon, down_remove = _enlarge_end(transcript,
                                                                               backup, end_transcript, up_exons, new_first_exon)
    except KeyError as e:
        logger.error(f"{e}")
    # first_exon, last_exon = transcript.exons[0], transcript.exons[-1]

    assert upstream >= 0 and downstream >= 0

    if transcript.monoexonic is False:
        if up_remove is True:
            # Remove the first exon
            up_exons = sorted(up_exons)
            transcript.exons[0] = up_exons[-1]
            up_exons.remove(up_exons[-1])
            # transcript.remove_exon(first_exon)
        if down_remove is True:  # (up_remove is False or first_exon != last_exon):
            down_exons = sorted(down_exons)
            transcript.exons[-1] = down_exons[-1]
            down_exons.remove(down_exons[-1])
        # transcript.remove_exon(last_exon)
    elif new_exon is not None:
        transcript.exons[0] = new_exon

    new_exons = up_exons + down_exons
    if not any([len(new_exons) > 0, up_remove, down_remove]):
        logger.debug("%s does not need to be expanded, exiting", transcript.id)
        return backup

    transcript.add_exons(new_exons)
    transcript.start, transcript.end = None, None
    transcript.finalize()

    if transcript.strand == "-":
        downstream, upstream = upstream, downstream

    if backup.is_coding:
        seq = check_expanded(transcript, backup, start_transcript, end_transcript,
                             fai, upstream, downstream, logger)
        transcript = enlarge_orfs(transcript, backup, seq, upstream, downstream, logger)
        transcript.finalize()

    logger.debug("%s: start (before %s, now %s, %s), end (before %s, now %s, %s)",
                 transcript.id,
                 backup.start, transcript.start, transcript.start < backup.start,
                 backup.end, transcript.end, transcript.end > backup.end)
    if transcript.start < backup.start or transcript.end > backup.end:
        transcript.attributes["padded"] = True

    # Now check that we have a valid expansion
    if backup.is_coding and not transcript.is_coding:
        # Something has gone wrong. Just return the original transcript.
        assert new_exons
        logger.info("Padding %s would lead to an invalid CDS (up exons: %s). Aborting.",
                    transcript.id, up_exons)
        return backup
    elif backup.is_coding:
        abort = False
        if backup.strand == "-" and backup.combined_cds_end < transcript.combined_cds_end:
            abort = True
        elif backup.strand != "-" and backup.combined_cds_end > transcript.combined_cds_end:
            abort = True
        if abort is True:
            msg = "Padding {} (strand: {}) would lead to an in-frame stop codon ({} to {}, \
vs original {} to {}. Aborting.".format(
                transcript.id, backup.strand, transcript.combined_cds_start, transcript.combined_cds_end,
                backup.combined_cds_start, backup.combined_cds_end)
            logger.info(msg)
            return backup

    return transcript


def _enlarge_start(transcript: Transcript,
                   backup: Transcript,
                   start_transcript: Transcript) -> (int, list, [None, tuple], bool):

    """This method will enlarge the transcript at the 5' end, using another transcript as the template.
    :param transcript: the original transcript to modify.
    :param backup: a copy of the transcript. As we are modifying the original one, we do need a hard copy.
    :param start_transcript: the template transcript.

    The function returns the following:
    :returns: the upstream modification, the list of upstream exons to add, the new first exon (if any),
              a boolean flag indicating whether the first exon of the transcript should be removed.
    """

    upstream = 0
    up_exons = []
    new_first_exon = None
    to_remove = False
    if start_transcript:
        transcript.start = start_transcript.start
        upstream_exons = sorted(
            [_ for _ in start_transcript.find_upstream(transcript.exons[0][0], transcript.exons[0][1])
             if _.value == "exon"])
        intersecting_upstream = sorted(start_transcript.search(
            transcript.exons[0][0], transcript.exons[0][1]))

        if not intersecting_upstream:
            raise KeyError("No exon or intron found to be intersecting with %s vs %s, this is a mistake",
                           transcript.id, start_transcript.id)

        if intersecting_upstream[0].value == "exon":
            new_first_exon = (min(intersecting_upstream[0][0], backup.start),
                              transcript.exons[0][1])
            if new_first_exon != transcript.exons[0]:
                upstream += backup.start - new_first_exon[0]
                up_exons.append(new_first_exon)
                to_remove = True
            else:
                new_first_exon = None
            if intersecting_upstream[0] in upstream_exons:
                upstream_exons.remove(intersecting_upstream[0])
            upstream += sum(_[1] - _[0] + 1 for _ in upstream_exons)
            up_exons.extend([(_[0], _[1]) for _ in upstream_exons])
        elif intersecting_upstream[0].value == "intron":
            # Check whether the first exon of the model *ends* within an *intron* of the template
            # If that is the case, we have to keep the first exon in place and
            # just expand it until the end
            # Now we have to expand until the first exon in the upstream_exons
            if intersecting_upstream[0][1] == transcript.exons[0][0] - 1:
                assert upstream_exons
                to_remove = False
            elif upstream_exons:
                to_remove = True
                upstream_exon = upstream_exons[-1]
                new_first_exon = (upstream_exon[0], transcript.exons[0][1])
                upstream_exons.remove(upstream_exon)
                upstream += backup.start - new_first_exon[0]
                up_exons.append(new_first_exon)
            else:
                # Something fishy going on here. Let us double check everything.
                if start_transcript.exons[0][0] == transcript.start:
                    raise ValueError(
                        "Something has gone wrong. The template transcript should have returned upstream exons."
                    )
                elif start_transcript.exons[0][0] < transcript.start:
                    raise ValueError(
                        "Something has gone wrong. We should have found the correct exons."
                    )
                else:
                    pass

            upstream += sum(_[1] - _[0] + 1 for _ in upstream_exons)
            up_exons.extend([(_[0], _[1]) for _ in upstream_exons])

    return upstream, up_exons, new_first_exon, to_remove


def _enlarge_end(transcript: Transcript,
                 backup: Transcript,
                 end_transcript: Transcript,
                 up_exons: list,
                 new_first_exon: [None, tuple]) -> [int, list, list, bool]:

    """
    This method will enlarge the transcript at the 5' end, using another transcript as the template.
    :param transcript: the original transcript to modify.
    :param backup: a copy of the transcript. As we are modifying the original one, we do need a hard copy.
    :param end_transcript: the template transcript.
    :param up_exons: the list of exons added at the 5' end.
    :param new_first_exon: the new coordinates of what used to be the first exon of the transcript.
                           This is necessary because if the transcript is monoexonic, we might need to re-modify it.

    The function returns the following:
    :returns: the downstream modification, the (potentially modified) list of upstream exons to add,
              the list of downstream exons to add, a boolean flag indicating whether the last exon of the transcript
              should be removed.
    """

    downstream = 0
    down_exons = []
    to_remove = False
    new_exon = new_first_exon[:] if new_first_exon is not None else None

    if end_transcript:
        transcript.end = end_transcript.end
        downstream_exons = sorted(
            [_ for _ in end_transcript.find_downstream(transcript.exons[-1][0], transcript.exons[-1][1])
             if _.value == "exon"])
        intersecting_downstream = sorted(end_transcript.search(
            transcript.exons[-1][0] + 1, transcript.exons[-1][1]))
        if not intersecting_downstream:
            raise KeyError("No exon or intron found to be intersecting with %s vs %s, this is a mistake",
                           transcript.id, end_transcript.id)
        # We are taking the right-most intersecting element.
        if intersecting_downstream[-1].value == "exon":
            if transcript.monoexonic and new_first_exon is not None:
                new_exon = (new_first_exon[0], max(intersecting_downstream[-1][1], new_first_exon[1]))
                if new_exon != new_first_exon:
                    up_exons.remove(new_first_exon)
                    downstream += new_exon[1] - backup.end
                    down_exons.append(new_exon)
                    to_remove = True
            else:
                new_exon = (transcript.exons[-1][0],
                            max(intersecting_downstream[-1][1], transcript.exons[-1][1]))
                if new_exon != transcript.exons[-1]:
                    downstream += new_exon[1] - backup.end
                    down_exons.append(new_exon)
                    to_remove = True

            if intersecting_downstream[-1] in downstream_exons:
                downstream_exons.remove(intersecting_downstream[-1])
            downstream += sum(_[1] - _[0] + 1 for _ in downstream_exons)
            down_exons.extend([(_[0], _[1]) for _ in downstream_exons])
        elif intersecting_downstream[-1].value == "intron":
            # Now we have to expand until the first exon in the upstream_exons
            if intersecting_downstream[-1][0] == transcript.exons[-1][1] + 1:
                assert downstream_exons
                to_remove = False
            elif downstream_exons:
                downstream_exon = downstream_exons[0]
                assert downstream_exon[1] > backup.end
                assert downstream_exon[0] > backup.end
                if transcript.monoexonic and new_first_exon is not None:
                    new_exon = (new_first_exon[0], downstream_exon[1])
                    up_exons.remove(new_first_exon)
                    to_remove = True
                else:
                    new_exon = (transcript.exons[-1][0], downstream_exon[1])
                    to_remove = True
                downstream_exons.remove(downstream_exon)
                downstream += new_exon[1] - backup.end
                down_exons.append(new_exon)
            else:
                # Something fishy going on here. Let us double check everything.
                if end_transcript.exons[-1][1] == transcript.end:
                    raise ValueError(
                        "Something has gone wrong. The template transcript should have returned upstream exons."
                    )
                elif end_transcript.exons[-1][1] > transcript.end:
                    raise ValueError(
                        "Something has gone wrong. We should have found the correct exons."
                    )
            downstream += sum(_[1] - _[0] + 1 for _ in downstream_exons)
            down_exons.extend([(_[0], _[1]) for _ in downstream_exons])

    return downstream, up_exons, down_exons, new_exon, to_remove


def check_expanded(transcript, backup, start_transcript, end_transcript, fai, upstream, downstream, logger) -> str:

    """
    This function checks that the expanded transcript is valid, and it also calculates and returns its cDNA sequence.
    :param transcript: the modified transcript.
    :param backup: The original transcript, before expansion.
    :param start_transcript: the transcript used as template at the 5' end.
    :param end_transcript: the transcript used as template at the 3' end.
    :param fai: The pysam.libcfaidx.FastaFile object indexing the genome.
    :param upstream: the amount of transcriptomic base-pairs added to the transcript at its 5' end.
    :param downstream: the amount of transcriptomic base-pairs added to the transcript at its 3' end.
    :param logger: the logger to use.
    :returns: the cDNA of the modified transcript, as a standard Python string.
    """

    assert transcript.exons != backup.exons
    assert transcript.end <= fai.get_reference_length(transcript.chrom), (
        transcript.end, fai.get_reference_length(transcript.chrom))
    genome_seq = fai.fetch(transcript.chrom, transcript.start - 1, transcript.end)

    if not (transcript.exons[-1][1] - transcript.start + 1 == len(genome_seq)):
        error = "{} should have a sequence of length {} ({} start, {} end), but one of length {} has been given"
        error = error.format(transcript.id, transcript.exons[-1][1] - transcript.start + 1,
                             transcript.start, transcript.end, len(genome_seq))
        logger.error(error)
        raise InvalidTranscript(error)
    seq = TranscriptChecker(transcript, genome_seq, is_reference=True).cdna
    assert len(seq) == transcript.cdna_length, (len(seq), transcript.cdna_length, transcript.exons)
    if not len(seq) == backup.cdna_length + upstream + downstream:
        error = [len(seq), backup.cdna_length + upstream + downstream,
                 backup.cdna_length, upstream, downstream,
                 (transcript.start, transcript.end), (backup.id, backup.start, backup.end),
                 (None if not start_transcript else (start_transcript.id, (start_transcript.start,
                                                                           start_transcript.end))),
                 (None if not end_transcript else (end_transcript.id, (end_transcript.start,
                                                                       end_transcript.end))),
                 (backup.id, backup.exons),
                 None if not start_transcript else (start_transcript.id, start_transcript.exons),
                 None if not end_transcript else (end_transcript.id, end_transcript.exons),
                 (transcript.id + "_expanded", transcript.exons),
                 set.difference(set(transcript.exons), set(backup.exons)),
                 set.difference(set(backup.exons), set(transcript.exons))
                 ]
        error = "\n".join([str(_) for _ in error])
        raise AssertionError(error)
    return seq


def enlarge_orfs(transcript: Transcript,
                 backup: Transcript,
                 seq: str,
                 upstream: int,
                 downstream: int,
                 logger) -> Transcript:

    """
    This method will take an expanded transcript and recalculate its ORF(s). As a consequence of the expansion,
    truncated transcripts might become whole.
    :param transcript: the expanded transcript.
    :param backup: the original transcript. Used to extract the original ORF(s).
    :param seq: the new cDNA sequence of the expanded transcript.
    :param upstream: the amount of expansion that happened at the 5'.
    :param downstream: the amount of expansion that happened at the 3'.
    :param logger: the logger.
    :returns: the modified transcript with the ORF(s) recalculated.
    """

    if backup.combined_cds_length > 0:
        try:
            internal_orfs = list(backup.get_internal_orf_beds())
        except (ValueError, TypeError, AssertionError):
            logger.error("Something went wrong with the CDS extraction for %s. Stripping it.",
                         backup.id)
            internal_orfs = []
    else:
        internal_orfs = []

    if not internal_orfs:
        return transcript

    new_orfs = []
    for orf in internal_orfs:
        logger.debug("Old ORF: %s", str(orf))
        try:
            logger.debug("Sequence for %s: %s[..]%s (upstream %s, downstream %s)",
                         transcript.id, seq[:10], seq[-10:], upstream, downstream)
            orf.expand(seq, upstream, downstream, expand_orf=True, logger=logger)
        except AssertionError as err:
            logger.error(err)
            logger.error("%s, %s, %s, %s",
                         upstream,
                         downstream,
                         transcript.exons,
                         transcript.cdna_length)
            raise AssertionError(err)
        logger.debug("New ORF: %s", str(orf))
        if orf.coding is False:
            raise ValueError(orf)
        elif orf.invalid:
            raise InvalidTranscript(orf.invalid_reason)

        new_orfs.append(orf)

    transcript.load_orfs(new_orfs)
    transcript.finalize()
    if backup.is_coding and not transcript.is_coding:
        raise InvalidTranscript(new_orfs)
    return transcript



#!/usr/bin/env python3

import argparse
import sys
import pysam
from Mikado.transcripts.transcript import Transcript
from collections import Counter


def to_bam(string):
    return pysam.AlignmentFile(string, mode="rb")


def main():
    parser = argparse.ArgumentParser("Script to convert from BAM to GTF, for PB alignments")
    parser.add_argument("--strict", action="store_true", default=False,
                        help="Switch. If set, this script will never output multiexonic transcripts \
                        without a defined strand.")
    parser.add_argument("--outfmt", choices=["gtf", "bed12"], default="gtf")
    parser.add_argument("bam", type=to_bam, help="Input BAM file")
    parser.add_argument("out", nargs="?", default=sys.stdout, type=argparse.FileType("wt"),
                        help="Optional output file")
    args = parser.parse_args()

    # M 0 alignment match (can be a sequence match or mismatch)
    # I 1 insertion to the reference
    # D 2 deletion from the reference
    # N 3 skipped region from the reference
    # S 4 soft clipping (clipped sequences present in SEQ)
    # H 5 hard clipping (clipped sequences NOT present in SEQ)
    # P 6 padding (silent deletion from padded reference)
    # = 7 sequence match
    # X 8 sequence mismatch

    name_counter = Counter()

    for record in args.bam:
        if record.is_unmapped is True:
            continue
        transcript = Transcript(record, accept_undefined_multi=(not args.strict))
        if name_counter.get(record.query_name):
            name = "{}_{}".format(record.query_name, name_counter.get(record.query_name))
        else:
            name = record.query_name

        if name != transcript.id:
            transcript.alias = transcript.id
            transcript.id = name

        transcript.parent = transcript.attributes["gene_id"] = "{0}.gene".format(name)
        name_counter.update([record.query_name])
        transcript.source = "bam2gtf"
        print(transcript.format(args.outfmt), file=args.out)


main()
